import numpy as np
import pytest

import autolens as al


@pytest.fixture(name="grid")
def make_grid():
    mask = al.Mask(
        np.array([[True, False, True], [False, False, False], [True, False, True]]),
        pixel_scale=1.0,
        sub_size=1,
    )

    return al.Grid.from_mask(mask=mask)


class TestGrid:
    def test__from_mask__compare_to_array_util(self):
        mask = np.array(
            [
                [True, True, False, False],
                [True, False, True, True],
                [True, True, False, False],
            ]
        )
        mask = al.Mask(array=mask, pixel_scale=2.0, sub_size=1)

        grid_via_util = al.grid_util.grid_1d_from_mask_pixel_scales_sub_size_and_origin(
            mask=mask, sub_size=1, pixel_scales=(2.0, 2.0)
        )

        grid = al.Grid.from_mask(mask=mask)

        assert type(grid) == al.Grid
        assert grid == pytest.approx(grid_via_util, 1e-4)
        assert grid.pixel_scale == 2.0
        assert (
            grid.mask.mapping.mask_1d_index_tomask_index
            == mask.mapping.mask_1d_index_tomask_index
        ).all()
        assert grid.interpolator == None

        grid_2d = mask.mapping.grid_2d_from_grid_1d(grid_1d=grid)

        assert (grid.in_2d == grid_2d).all()

        mask = np.array([[True, True, True], [True, False, False], [True, True, False]])

        mask = al.Mask(mask, pixel_scale=3.0, sub_size=2)

        grid_via_util = al.grid_util.grid_1d_from_mask_pixel_scales_sub_size_and_origin(
            mask=mask, pixel_scales=(3.0, 3.0), sub_size=2
        )

        grid = al.Grid.from_mask(mask=mask)

        assert grid == pytest.approx(grid_via_util, 1e-4)

    def test__unlensed_unsubbed_grid_property_is_grid_with_sub_size_1(self):
        mask = np.array(
            [
                [True, True, False, False],
                [True, False, True, True],
                [True, True, False, False],
            ]
        )
        mask = al.Mask(array=mask, pixel_scale=2.0, sub_size=2)

        grid = al.Grid.from_mask(mask=mask)

        assert (
            grid.unlensed_unsubbed_1d
            == np.array([[2.0, 1.0], [2.0, 3.0], [0.0, -1.0], [-2.0, 1.0], [-2.0, 3.0]])
        ).all()

    def test__grid_unlensed_1d_property__compare_to_grid_util(self):
        mask = np.array(
            [
                [True, True, False, False],
                [True, False, True, True],
                [True, True, False, False],
            ]
        )

        mask = al.Mask(array=mask, pixel_scale=2.0, sub_size=1)

        grid = al.Grid(arr=np.array([[1.0, 1.0], [1.0, 1.0]]), mask=mask)

        grid_via_util = al.grid_util.grid_1d_from_mask_pixel_scales_sub_size_and_origin(
            mask=mask, sub_size=1, pixel_scales=(2.0, 2.0)
        )

        assert grid.unlensed_1d == pytest.approx(grid_via_util, 1e-4)

        grid_via_util = al.grid_util.grid_1d_from_shape_pixel_scales_sub_size_and_origin(
            shape=(3, 4), sub_size=1, pixel_scales=(2.0, 2.0)
        )

        assert grid.unlensed_unmasked_1d == pytest.approx(grid_via_util, 1e-4)

        mask = al.Mask(
            np.array([[True, False, True], [False, False, False], [True, False, True]]),
            pixel_scale=1.0,
            sub_size=1,
        )

        grid = al.Grid.from_mask(mask=mask)

        assert grid.unlensed_1d == pytest.approx(
            np.array([[1, 0], [0, -1], [0, 0], [0, 1], [-1, 0]]), 1e-4
        )

        grid_via_util = al.grid_util.grid_1d_from_mask_pixel_scales_sub_size_and_origin(
            mask=np.full((3, 3), False), pixel_scales=(1.0, 1.0), sub_size=1
        )

        assert grid.unlensed_unmasked_1d == pytest.approx(grid_via_util, 1e-4)

        mask = al.Mask(
            np.array([[True, False, True], [False, False, False], [True, False, True]]),
            pixel_scale=1.0,
            sub_size=2,
        )

        grid = al.Grid.from_mask(mask=mask)

        grid_via_util = al.grid_util.grid_1d_from_mask_pixel_scales_sub_size_and_origin(
            mask=np.full((3, 3), False), pixel_scales=(1.0, 1.0), sub_size=2
        )

        assert grid.unlensed_unmasked_1d == pytest.approx(grid_via_util, 1e-4)

    def test__grid_unlensed_2d_property__compare_to_grid_util(self):
        mask = np.array(
            [
                [True, True, False, False],
                [True, False, True, True],
                [True, True, False, False],
            ]
        )

        mask = al.Mask(array=mask, pixel_scale=2.0, sub_size=1)

        grid = al.Grid(arr=np.array([[1.0, 1.0], [1.0, 1.0]]), mask=mask)

        grid_via_util = al.grid_util.grid_2d_from_mask_pixel_scales_sub_size_and_origin(
            mask=mask, sub_size=1, pixel_scales=(2.0, 2.0)
        )

        assert grid.unlensed_2d == pytest.approx(grid_via_util, 1e-4)

        grid_via_util = al.grid_util.grid_2d_from_shape_pixel_scales_sub_size_and_origin(
            shape=(3, 4), sub_size=1, pixel_scales=(2.0, 2.0)
        )

        assert grid.unlensed_unmasked_2d == pytest.approx(grid_via_util, 1e-4)

        mask = al.Mask(
            np.array([[True, False, True], [False, False, False], [True, False, True]]),
            pixel_scale=1.0,
            sub_size=1,
        )

        grid = al.Grid.from_mask(mask=mask)

        assert grid.unlensed_2d == pytest.approx(
            np.array(
                [
                    [[0, 0], [1, 0], [0, 0]],
                    [[0, -1], [0, 0], [0, 1]],
                    [[0, 0], [-1, 0], [0, 0]],
                ]
            ),
            1e-4,
        )

        grid_via_util = al.grid_util.grid_2d_from_mask_pixel_scales_sub_size_and_origin(
            mask=np.full((3, 3), False), pixel_scales=(1.0, 1.0), sub_size=1
        )

        assert grid.unlensed_unmasked_2d == pytest.approx(grid_via_util, 1e-4)

        mask = al.Mask(
            np.array([[True, False, True], [False, False, False], [True, False, True]]),
            pixel_scale=1.0,
            sub_size=2,
        )

        grid = al.Grid.from_mask(mask)

        grid_via_util = al.grid_util.grid_2d_from_mask_pixel_scales_sub_size_and_origin(
            mask=np.full((3, 3), False), pixel_scales=(1.0, 1.0), sub_size=2
        )

        assert grid.unlensed_unmasked_2d == pytest.approx(grid_via_util, 1e-4)

    def test__from_shape_and_pixel_scale__compare_to_grid_util(self):
        mask = np.array(
            [
                [False, False, False, False],
                [False, False, False, False],
                [False, False, False, False],
            ]
        )
        mask = al.Mask(array=mask, pixel_scale=2.0, sub_size=1)

        grid_via_util = al.grid_util.grid_1d_from_mask_pixel_scales_sub_size_and_origin(
            mask=mask, pixel_scales=(2.0, 2.0), sub_size=1
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_size(
            shape=(3, 4), pixel_scale=2.0, sub_size=1
        )

        assert type(grid) == al.Grid
        assert grid == pytest.approx(grid_via_util, 1e-4)
        assert grid.pixel_scale == 2.0
        assert (
            grid.mapping.mask_1d_index_tomask_index
            == mask.mapping.mask_1d_index_tomask_index
        ).all()

        mask = np.array(
            [[False, False, False], [False, False, False], [False, False, False]]
        )

        grid_via_util = al.grid_util.grid_1d_from_mask_pixel_scales_sub_size_and_origin(
            mask=mask, pixel_scales=(3.0, 3.0), sub_size=2
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_size(
            shape=(3, 3), pixel_scale=3.0, sub_size=2
        )

        assert grid == pytest.approx(grid_via_util, 1e-4)

    def test__blurring_grid_from_mask__compare_to_array_util(self):
        mask = np.array(
            [
                [True, True, True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True, True, True],
                [True, True, True, True, False, True, True, True, True],
                [True, True, True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True, True, True],
            ]
        )

        mask = al.Mask(array=mask, pixel_scale=2.0, sub_size=2)

        blurring_mask_util = al.mask_util.blurring_mask_from_mask_and_psf_shape(
            mask=mask, psf_shape=(3, 5)
        )

        blurring_grid_util = al.grid_util.grid_1d_from_mask_pixel_scales_sub_size_and_origin(
            mask=blurring_mask_util, pixel_scales=(2.0, 2.0), sub_size=1
        )

        grid = al.Grid.from_mask(mask=mask)

        blurring_grid = grid.blurring_grid_from_psf_shape(psf_shape=(3, 5))

        blurring_mask = mask.blurring_mask_from_psf_shape(psf_shape=(3, 5))

        assert blurring_grid == pytest.approx(blurring_grid_util, 1e-4)
        assert blurring_grid.pixel_scale == 2.0
        assert (
            blurring_grid.mask.mapping.mask_1d_index_tomask_index
            == blurring_mask.mapping.mask_1d_index_tomask_index
        ).all()
        assert blurring_grid.sub_size == 1

    def test__blurring_grid_from_psf_shape__compare_to_array_util(self):
        mask = np.array(
            [
                [True, True, True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True, True, True],
                [True, True, False, True, True, True, False, True, True],
                [True, True, True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True, True, True],
                [True, True, False, True, True, True, False, True, True],
                [True, True, True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True, True, True],
            ]
        )

        mask = al.Mask(array=mask, pixel_scale=2.0, sub_size=2)

        blurring_mask_util = al.mask_util.blurring_mask_from_mask_and_psf_shape(
            mask=mask, psf_shape=(3, 5)
        )

        blurring_grid_util = al.grid_util.grid_1d_from_mask_pixel_scales_sub_size_and_origin(
            mask=blurring_mask_util, pixel_scales=(2.0, 2.0), sub_size=1
        )

        mask = al.Mask(array=mask, pixel_scale=2.0, sub_size=2)
        blurring_grid = al.Grid.blurring_grid_from_mask_and_psf_shape(
            mask=mask, psf_shape=(3, 5)
        )

        blurring_mask = mask.blurring_mask_from_psf_shape(psf_shape=(3, 5))

        assert blurring_grid == pytest.approx(blurring_grid_util, 1e-4)
        assert blurring_grid.pixel_scale == 2.0
        assert (
            blurring_grid.mask.mapping.mask_1d_index_tomask_index
            == blurring_mask.mapping.mask_1d_index_tomask_index
        ).all()
        assert blurring_grid.sub_size == 1

    def test__padded_grid_from_psf_shape__matches_grid_2d_after_padding(self):

        grid = al.Grid.from_shape_pixel_scale_and_sub_size(
            shape=(4, 4), pixel_scale=3.0, sub_size=1
        )

        padded_grid = grid.padded_grid_from_psf_shape(psf_shape=(3, 3))

        padded_grid_util = al.grid_util.grid_1d_from_mask_pixel_scales_sub_size_and_origin(
            mask=np.full((6, 6), False), pixel_scales=(3.0, 3.0), sub_size=1
        )

        assert padded_grid.shape == (36, 2)
        assert (padded_grid.mask == np.full(fill_value=False, shape=(6, 6))).all()
        assert (padded_grid == padded_grid_util).all()
        assert padded_grid.interpolator is None

        grid = al.Grid.from_shape_pixel_scale_and_sub_size(
            shape=(4, 5), pixel_scale=2.0, sub_size=1
        )

        padded_grid = grid.padded_grid_from_psf_shape(psf_shape=(3, 3))

        padded_grid_util = al.grid_util.grid_1d_from_mask_pixel_scales_sub_size_and_origin(
            mask=np.full((6, 7), False), pixel_scales=(2.0, 2.0), sub_size=1
        )

        assert padded_grid.shape == (42, 2)
        assert (padded_grid == padded_grid_util).all()

        grid = al.Grid.from_shape_pixel_scale_and_sub_size(
            shape=(5, 4), pixel_scale=1.0, sub_size=1
        )

        padded_grid = grid.padded_grid_from_psf_shape(psf_shape=(3, 3))

        padded_grid_util = al.grid_util.grid_1d_from_mask_pixel_scales_sub_size_and_origin(
            mask=np.full((7, 6), False), pixel_scales=(1.0, 1.0), sub_size=1
        )

        assert padded_grid.shape == (42, 2)
        assert (padded_grid == padded_grid_util).all()

        grid = al.Grid.from_shape_pixel_scale_and_sub_size(
            shape=(5, 5), pixel_scale=8.0, sub_size=1
        )

        padded_grid = grid.padded_grid_from_psf_shape(psf_shape=(2, 5))

        padded_grid_util = al.grid_util.grid_1d_from_mask_pixel_scales_sub_size_and_origin(
            mask=np.full((6, 9), False), pixel_scales=(8.0, 8.0), sub_size=1
        )

        assert padded_grid.shape == (54, 2)
        assert (padded_grid == padded_grid_util).all()

        mask = al.Mask(array=np.full((5, 4), False), pixel_scale=2.0, sub_size=2)

        grid = al.Grid.from_mask(mask=mask)

        padded_grid = grid.padded_grid_from_psf_shape(psf_shape=(3, 3))

        padded_grid_util = al.grid_util.grid_1d_from_mask_pixel_scales_sub_size_and_origin(
            mask=np.full((7, 6), False), pixel_scales=(2.0, 2.0), sub_size=2
        )

        assert padded_grid.shape == (168, 2)
        assert (padded_grid.mask == np.full(fill_value=False, shape=(7, 6))).all()
        assert padded_grid == pytest.approx(padded_grid_util, 1e-4)
        assert padded_grid.interpolator is None

        mask = al.Mask(array=np.full((2, 5), False), pixel_scale=8.0, sub_size=4)

        grid = al.Grid.from_mask(mask=mask)

        padded_grid = grid.padded_grid_from_psf_shape(psf_shape=(5, 5))

        padded_grid_util = al.grid_util.grid_1d_from_mask_pixel_scales_sub_size_and_origin(
            mask=np.full((6, 9), False), pixel_scales=(8.0, 8.0), sub_size=4
        )

        assert padded_grid.shape == (864, 2)
        assert (padded_grid.mask == np.full(fill_value=False, shape=(6, 9))).all()
        assert padded_grid == pytest.approx(padded_grid_util, 1e-4)

    def test__padded_grid_from_psf_shape__has_interpolator_grid_if_had_one_before(self):
        grid = al.Grid.from_shape_pixel_scale_and_sub_size(
            shape=(4, 4), pixel_scale=3.0, sub_size=1
        )

        grid = grid.new_grid_with_interpolator(pixel_scale_interpolation_grid=0.1)

        padded_grid = grid.padded_grid_from_psf_shape(psf_shape=(3, 3))

        assert padded_grid.interpolator is not None
        assert padded_grid.interpolator.pixel_scale_interpolation_grid == 0.1

        mask = al.Mask.unmasked_from_shape_pixel_scale_and_sub_size(
            shape=(6, 6), pixel_scale=3.0, sub_size=1
        )

        interpolator = al.Interpolator.from_mask_grid_and_pixel_scale_interpolation_grids(
            mask=mask, grid=padded_grid, pixel_scale_interpolation_grid=0.1
        )

        assert (padded_grid.interpolator.vtx == interpolator.vtx).all()
        assert (padded_grid.interpolator.wts == interpolator.wts).all()

        mask = al.Mask(array=np.full((5, 4), False), pixel_scale=2.0, sub_size=2)

        grid = al.Grid.from_mask(mask=mask)

        grid = grid.new_grid_with_interpolator(pixel_scale_interpolation_grid=0.1)

        padded_grid = grid.padded_grid_from_psf_shape(psf_shape=(3, 3))

        assert padded_grid.interpolator is not None
        assert padded_grid.interpolator.pixel_scale_interpolation_grid == 0.1

        mask = al.Mask.unmasked_from_shape_pixel_scale_and_sub_size(
            shape=(7, 6), pixel_scale=2.0, sub_size=2
        )

        interpolator = al.Interpolator.from_mask_grid_and_pixel_scale_interpolation_grids(
            mask=mask, grid=padded_grid, pixel_scale_interpolation_grid=0.1
        )

        assert (padded_grid.interpolator.vtx == interpolator.vtx).all()
        assert (padded_grid.interpolator.wts == interpolator.wts).all()

    def test__trimmed_array_2d_from_padded_array_1d_and_image_shape(self):
        mask = al.Mask(array=np.full((4, 4), False), pixel_scale=1.0, sub_size=1)

        grid = al.Grid(arr=np.empty((0)), mask=mask)

        array_1d = np.array(
            [
                1.0,
                2.0,
                3.0,
                4.0,
                5.0,
                6.0,
                7.0,
                8.0,
                9.0,
                1.0,
                2.0,
                3.0,
                4.0,
                5.0,
                6.0,
                7.0,
            ]
        )

        array_2d = grid.trimmed_array_2d_from_padded_array_1d_and_image_shape(
            padded_array_1d=array_1d, image_shape=(2, 2)
        )

        assert (array_2d == np.array([[6.0, 7.0], [1.0, 2.0]])).all()

        mask = al.Mask(array=np.full((5, 3), False), pixel_scale=1.0, sub_size=1)

        grid = al.Grid(arr=np.empty((0)), mask=mask)

        array_1d = np.array(
            [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0]
        )

        array_2d = grid.trimmed_array_2d_from_padded_array_1d_and_image_shape(
            padded_array_1d=array_1d, image_shape=(3, 1)
        )

        assert (array_2d == np.array([[5.0], [8.0], [2.0]])).all()

        mask = al.Mask(array=np.full((3, 5), False), pixel_scale=1.0, sub_size=1)

        grid = al.Grid(arr=np.empty((0)), mask=mask)

        array_1d = np.array(
            [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0]
        )

        array_2d = grid.trimmed_array_2d_from_padded_array_1d_and_image_shape(
            padded_array_1d=array_1d, image_shape=(1, 3)
        )

        assert (array_2d == np.array([[7.0, 8.0, 9.0]])).all()

    def test_sub_mask_1d_index_to_mask_1d_index(self, grid):
        assert (
            grid.mapping.sub_mask_1d_index_to_mask_1d_index == np.array(range(5))
        ).all()

    def test__sub_border_1d_indexes__compare_to_array_util(self):
        mask = np.array(
            [
                [False, False, False, False, False, False, False, True],
                [False, True, True, True, True, True, False, True],
                [False, True, False, False, False, True, False, True],
                [False, True, False, True, False, True, False, True],
                [False, True, False, False, False, True, False, True],
                [False, True, True, True, True, True, False, True],
                [False, False, False, False, False, False, False, True],
            ]
        )

        mask = al.Mask(array=mask, pixel_scale=2.0, sub_size=2)

        sub_border_1d_indexes_util = al.mask_util.sub_border_pixel_1d_indexes_from_mask_and_sub_size(
            mask=mask, sub_size=2
        )

        grid = al.Grid.from_mask(mask=mask)

        assert grid.sub_border_1d_indexes == pytest.approx(
            sub_border_1d_indexes_util, 1e-4
        )

    def test__masked_shape_arcsec(self):
        mask = al.Mask.circular(
            shape=(3, 3), radius_arcsec=1.0, pixel_scale=1.0, sub_size=1
        )

        grid = al.Grid(arr=np.array([[1.5, 1.0], [-1.5, -1.0]]), mask=mask)
        assert grid.masked_shape_arcsec == (3.0, 2.0)

        grid = al.Grid(arr=np.array([[1.5, 1.0], [-1.5, -1.0], [0.1, 0.1]]), mask=mask)
        assert grid.masked_shape_arcsec == (3.0, 2.0)

        grid = al.Grid(arr=np.array([[1.5, 1.0], [-1.5, -1.0], [3.0, 3.0]]), mask=mask)
        assert grid.masked_shape_arcsec == (4.5, 4.0)

        grid = al.Grid(
            arr=np.array([[1.5, 1.0], [-1.5, -1.0], [3.0, 3.0], [7.0, -5.0]]), mask=mask
        )
        assert grid.masked_shape_arcsec == (8.5, 8.0)

    def test__in_radians(self):
        mask = np.array(
            [
                [True, True, False, False],
                [True, False, True, True],
                [True, True, False, False],
            ]
        )
        mask = al.Mask(array=mask, pixel_scale=2.0, sub_size=1)

        grid = al.Grid.from_mask(mask=mask)

        assert grid.in_radians[0, 0] == pytest.approx(0.00000969627362, 1.0e-8)
        assert grid.in_radians[0, 1] == pytest.approx(0.00000484813681, 1.0e-8)

        assert grid.in_radians[0, 0] == pytest.approx(
            2.0 * np.pi / (180 * 3600), 1.0e-8
        )
        assert grid.in_radians[0, 1] == pytest.approx(
            1.0 * np.pi / (180 * 3600), 1.0e-8
        )

    def test__yticks(self):
        mask = al.Mask.circular(
            shape=(3, 3), radius_arcsec=1.0, pixel_scale=1.0, sub_size=1
        )

        grid = al.Grid(arr=np.array([[1.5, 1.0], [-1.5, -1.0]]), mask=mask)
        assert grid.yticks == pytest.approx(np.array([-1.5, -0.5, 0.5, 1.5]), 1e-3)

        grid = al.Grid(arr=np.array([[3.0, 1.0], [-3.0, -1.0]]), mask=mask)
        assert grid.yticks == pytest.approx(np.array([-3.0, -1, 1.0, 3.0]), 1e-3)

        grid = al.Grid(arr=np.array([[5.0, 3.5], [2.0, -1.0]]), mask=mask)
        assert grid.yticks == pytest.approx(np.array([2.0, 3.0, 4.0, 5.0]), 1e-3)

    def test__xticks(self):
        mask = al.Mask.circular(
            shape=(3, 3), radius_arcsec=1.0, pixel_scale=1.0, sub_size=1
        )

        grid = al.Grid(arr=np.array([[1.0, 1.5], [-1.0, -1.5]]), mask=mask)
        assert grid.xticks == pytest.approx(np.array([-1.5, -0.5, 0.5, 1.5]), 1e-3)

        grid = al.Grid(arr=np.array([[1.0, 3.0], [-1.0, -3.0]]), mask=mask)
        assert grid.xticks == pytest.approx(np.array([-3.0, -1, 1.0, 3.0]), 1e-3)

        grid = al.Grid(arr=np.array([[3.5, 2.0], [-1.0, 5.0]]), mask=mask)
        assert grid.xticks == pytest.approx(np.array([2.0, 3.0, 4.0, 5.0]), 1e-3)

    def test__new_grid__with_interpolator__returns_grid_with_interpolator(self):
        mask = np.array(
            [
                [True, True, False, False],
                [True, False, True, True],
                [True, True, False, False],
            ]
        )
        mask = al.Mask(array=mask, pixel_scale=2.0, sub_size=1)

        grid = al.Grid.from_mask(mask=mask)

        grid_with_interp = grid.new_grid_with_interpolator(
            pixel_scale_interpolation_grid=1.0
        )

        assert (grid[:, :] == grid_with_interp[:, :]).all()
        assert grid.mask == grid_with_interp.mask

        interpolator_manual = al.Interpolator.from_mask_grid_and_pixel_scale_interpolation_grids(
            mask=mask, grid=grid, pixel_scale_interpolation_grid=1.0
        )

        assert (grid.interpolator.vtx == interpolator_manual.vtx).all()
        assert (grid.interpolator.wts == interpolator_manual.wts).all()

    def test__new_grid__with_binned__returns_grid_with_binned(self):
        mask = np.array(
            [
                [True, True, False, False],
                [True, False, True, True],
                [True, True, False, False],
            ]
        )
        mask = al.Mask(array=mask, pixel_scale=2.0, sub_size=1)

        grid = al.Grid.from_mask(mask=mask)

        grid_binned = grid.new_grid_with_binned_grid(binned_grid=1)

        assert grid.binned == 1


class TestGridBorder(object):
    def test__border_grid_for_simple_mask(self):
        mask = np.array(
            [
                [False, False, False, False, False, False, False, True],
                [False, True, True, True, True, True, False, True],
                [False, True, False, False, False, True, False, True],
                [False, True, False, True, False, True, False, True],
                [False, True, False, False, False, True, False, True],
                [False, True, True, True, True, True, False, True],
                [False, False, False, False, False, False, False, True],
            ]
        )

        mask = al.Mask(array=mask, pixel_scale=2.0, sub_size=2)

        grid = al.Grid.from_mask(mask=mask)

        assert (
            grid.sub_border_grid
            == np.array(
                [
                    [6.5, -7.5],
                    [6.5, -5.5],
                    [6.5, -3.5],
                    [6.5, -0.5],
                    [6.5, 1.5],
                    [6.5, 3.5],
                    [6.5, 5.5],
                    [4.5, -7.5],
                    [4.5, 5.5],
                    [2.5, -7.5],
                ]
            )
        ).all()

    def test__inside_border_no_relocations(self):
        mask = al.Mask.circular(
            shape=(30, 30), radius_arcsec=1.0, pixel_scale=0.1, sub_size=1
        )

        grid = al.Grid.from_mask(mask=mask)

        grid_to_relocate = al.Grid(
            arr=np.array([[0.1, 0.1], [0.3, 0.3], [-0.1, -0.2]]), mask=mask
        )

        relocated_grid = grid.relocated_grid_from_grid(grid=grid_to_relocate)

        assert (
            relocated_grid == np.array([[0.1, 0.1], [0.3, 0.3], [-0.1, -0.2]])
        ).all()
        assert relocated_grid.mask == mask
        assert relocated_grid.sub_size == 1

        mask = al.Mask.circular(
            shape=(30, 30), radius_arcsec=1.0, pixel_scale=0.1, sub_size=2
        )

        grid = al.Grid.from_mask(mask=mask)

        grid_to_relocate = al.Grid(
            arr=np.array([[0.1, 0.1], [0.3, 0.3], [-0.1, -0.2]]), mask=mask
        )

        relocated_grid = grid.relocated_grid_from_grid(grid=grid_to_relocate)

        assert (
            relocated_grid == np.array([[0.1, 0.1], [0.3, 0.3], [-0.1, -0.2]])
        ).all()
        assert relocated_grid.mask == mask
        assert relocated_grid.sub_size == 2

    def test__outside_border_are_relocations(self):
        mask = al.Mask.circular(
            shape=(30, 30), radius_arcsec=1.0, pixel_scale=0.1, sub_size=1
        )

        grid = al.Grid.from_mask(mask=mask)

        grid_to_relocate = al.Grid(
            arr=np.array([[10.1, 0.0], [0.0, 10.1], [-10.1, -10.1]]), mask=mask
        )

        relocated_grid = grid.relocated_grid_from_grid(grid=grid_to_relocate)

        assert relocated_grid == pytest.approx(
            np.array([[0.95, 0.0], [0.0, 0.95], [-0.7017, -0.7017]]), 0.1
        )
        assert relocated_grid.mask == mask
        assert relocated_grid.sub_size == 1

        mask = al.Mask.circular(
            shape=(30, 30), radius_arcsec=1.0, pixel_scale=0.1, sub_size=2
        )

        grid = al.Grid.from_mask(mask=mask)

        grid_to_relocate = al.Grid(
            arr=np.array([[10.1, 0.0], [0.0, 10.1], [-10.1, -10.1]]), mask=mask
        )

        relocated_grid = grid.relocated_grid_from_grid(grid=grid_to_relocate)

        assert relocated_grid == pytest.approx(
            np.array([[0.9778, 0.0], [0.0, 0.97788], [-0.7267, -0.7267]]), 0.1
        )
        assert relocated_grid.mask == mask
        assert relocated_grid.sub_size == 2

    def test__outside_border_are_relocations__positive_origin_included_in_relocate(
        self
    ):
        mask = al.Mask.circular(
            shape=(60, 60),
            radius_arcsec=1.0,
            pixel_scale=0.1,
            centre=(1.0, 1.0),
            sub_size=1,
        )

        grid = al.Grid.from_mask(mask=mask)

        grid_to_relocate = al.Grid(
            arr=np.array([[11.1, 1.0], [1.0, 11.1], [-11.1, -11.1]]),
            sub_size=1,
            mask=mask,
        )

        relocated_grid = grid.relocated_grid_from_grid(grid=grid_to_relocate)

        assert relocated_grid == pytest.approx(
            np.array(
                [[2.0, 1.0], [1.0, 2.0], [1.0 - np.sqrt(2) / 2, 1.0 - np.sqrt(2) / 2]]
            ),
            0.1,
        )
        assert relocated_grid.mask == mask
        assert relocated_grid.sub_size == 1

        mask = al.Mask.circular(
            shape=(60, 60),
            radius_arcsec=1.0,
            pixel_scale=0.1,
            centre=(1.0, 1.0),
            sub_size=2,
        )

        grid = al.Grid.from_mask(mask=mask)

        grid_to_relocate = al.Grid(
            arr=np.array([[11.1, 1.0], [1.0, 11.1], [-11.1, -11.1]]), mask=mask
        )

        relocated_grid = grid.relocated_grid_from_grid(grid=grid_to_relocate)

        assert relocated_grid == pytest.approx(
            np.array(
                [
                    [1.9263, 1.0 - 0.0226],
                    [1.0 - 0.0226, 1.9263],
                    [1.0 - 0.7267, 1.0 - 0.7267],
                ]
            ),
            0.1,
        )
        assert relocated_grid.mask == mask
        assert relocated_grid.sub_size == 2


class TestBinnedGrid:
    def test__from_mask_and_pixel_scale_binned_grid__correct_binned_bin_up_calculated(
        self, mask_7x7, grid_7x7
    ):
        mask_7x7.pixel_scale = 1.0
        binned_grid = al.BinnedGrid.from_mask_and_pixel_scale_binned_grid(
            mask=mask_7x7, pixel_scale_binned_grid=1.0
        )

        assert (binned_grid == grid_7x7).all()
        assert (binned_grid.mask == mask_7x7).all()
        assert binned_grid.bin_up_factor == 1
        assert (
            binned_grid.binned_mask_1d_index_to_mask_1d_indexes
            == np.array([[0], [1], [2], [3], [4], [5], [6], [7], [8]])
        ).all()

        mask_7x7.pixel_scale = 1.0
        binned_grid = al.BinnedGrid.from_mask_and_pixel_scale_binned_grid(
            mask=mask_7x7, pixel_scale_binned_grid=1.9
        )

        assert binned_grid.bin_up_factor == 1
        assert (binned_grid.mask == mask_7x7).all()
        assert (
            binned_grid.binned_mask_1d_index_to_mask_1d_indexes
            == np.array([[0], [1], [2], [3], [4], [5], [6], [7], [8]])
        ).all()

        mask_7x7.pixel_scale = 1.0
        binned_grid = al.BinnedGrid.from_mask_and_pixel_scale_binned_grid(
            mask=mask_7x7, pixel_scale_binned_grid=2.0
        )

        assert binned_grid.bin_up_factor == 2
        assert (
            binned_grid.mask
            == np.array(
                [
                    [True, True, True, True],
                    [True, False, False, True],
                    [True, False, False, True],
                    [True, True, True, True],
                ]
            )
        ).all()

        assert (
            binned_grid
            == np.array([[1.0, -1.0], [1.0, 1.0], [-1.0, -1.0], [-1.0, 1.0]])
        ).all()
        assert (
            binned_grid.binned_mask_1d_index_to_mask_1d_indexes
            == np.array([[0, -1, -1, -1], [1, 2, -1, -1], [3, 6, -1, -1], [4, 5, 7, 8]])
        ).all()

        mask_7x7.pixel_scale = 2.0
        binned_grid = al.BinnedGrid.from_mask_and_pixel_scale_binned_grid(
            mask=mask_7x7, pixel_scale_binned_grid=1.0
        )

        assert binned_grid.bin_up_factor == 1


class TestPixelizationGrid:
    def test_pix_grid__attributes(self):
        pix_grid = al.PixelizationGrid(
            arr=np.array([[1.0, 1.0], [2.0, 2.0]]),
            mask_1d_index_to_nearest_pixelization_1d_index=np.array([0, 1]),
        )

        assert type(pix_grid) == al.PixelizationGrid
        assert (pix_grid == np.array([[1.0, 1.0], [2.0, 2.0]])).all()
        assert (
            pix_grid.mask_1d_index_to_nearest_pixelization_1d_index == np.array([0, 1])
        ).all()

    def test__from_unmasked_sparse_shape_and_grid(self):
        mask = al.Mask(
            array=np.array(
                [[True, False, True], [False, False, False], [True, False, True]]
            ),
            pixel_scale=0.5,
            sub_size=1,
        )

        grid = al.Grid.from_mask(mask=mask)

        sparse_to_grid = al.SparseToGrid.from_grid_and_unmasked_2d_grid_shape(
            unmasked_sparse_shape=(10, 10), grid=grid
        )

        pixelization_grid = al.PixelizationGrid.from_grid_and_unmasked_2d_grid_shape(
            unmasked_sparse_shape=(10, 10), grid=grid
        )

        assert (sparse_to_grid.sparse == pixelization_grid).all()
        assert (
            sparse_to_grid.mask_1d_index_to_sparse_1d_index
            == pixelization_grid.mask_1d_index_to_nearest_pixelization_1d_index
        ).all()


class TestSparseToGrid:
    class TestUnmaskedShape:
        def test__properties_consistent_with_mapping_util(self):
            mask = al.Mask(
                array=np.array(
                    [[True, False, True], [False, False, False], [True, False, True]]
                ),
                pixel_scale=0.5,
                sub_size=1,
            )

            grid = al.Grid.from_mask(mask=mask)

            sparse_to_grid = al.SparseToGrid.from_grid_and_unmasked_2d_grid_shape(
                unmasked_sparse_shape=(10, 10), grid=grid
            )

            unmasked_sparse_grid_util = al.grid_util.grid_1d_from_shape_pixel_scales_sub_size_and_origin(
                shape=(10, 10), pixel_scales=(0.15, 0.15), sub_size=1, origin=(0.0, 0.0)
            )

            unmasked_sparse_grid_pixel_centres = grid.mask.grid_arcsec_to_grid_pixel_centres(
                grid_arcsec=unmasked_sparse_grid_util
            )

            total_sparse_pixels = al.mask_util.total_sparse_pixels_from_mask(
                mask=mask,
                unmasked_sparse_grid_pixel_centres=unmasked_sparse_grid_pixel_centres,
            )

            regular_to_unmasked_sparse_util = al.grid_util.grid_arcsec_1d_to_grid_pixel_indexes_1d(
                grid_arcsec_1d=grid,
                shape=(10, 10),
                pixel_scales=(0.15, 0.15),
                origin=(0.0, 0.0),
            ).astype(
                "int"
            )

            sparse_to_unmasked_sparse_util = al.sparse_mapping_util.sparse_to_unmasked_sparse_from_mask_and_pixel_centres(
                total_sparse_pixels=total_sparse_pixels,
                mask=mask,
                unmasked_sparse_grid_pixel_centres=unmasked_sparse_grid_pixel_centres,
            ).astype(
                "int"
            )

            unmasked_sparse_to_sparse_util = al.sparse_mapping_util.unmasked_sparse_to_sparse_from_mask_and_pixel_centres(
                mask=mask,
                unmasked_sparse_grid_pixel_centres=unmasked_sparse_grid_pixel_centres,
                total_sparse_pixels=total_sparse_pixels,
            ).astype(
                "int"
            )

            mask_1d_index_to_sparse_1d_index_util = al.sparse_mapping_util.mask_1d_index_to_sparse_1d_index_from_sparse_mappings(
                regular_to_unmasked_sparse=regular_to_unmasked_sparse_util,
                unmasked_sparse_to_sparse=unmasked_sparse_to_sparse_util,
            )

            sparse_grid_util = al.sparse_mapping_util.sparse_grid_from_unmasked_sparse_grid(
                unmasked_sparse_grid=unmasked_sparse_grid_util,
                sparse_to_unmasked_sparse=sparse_to_unmasked_sparse_util,
            )

            assert (
                sparse_to_grid.mask_1d_index_to_sparse_1d_index
                == mask_1d_index_to_sparse_1d_index_util
            ).all()
            assert (sparse_to_grid.sparse == sparse_grid_util).all()

        def test__sparse_grid_overlaps_mask_perfectly__masked_pixels_in_masked_sparse_grid(
            self
        ):
            mask = al.Mask(
                array=np.array(
                    [[True, False, True], [False, False, False], [True, False, True]]
                ),
                pixel_scale=1.0,
                sub_size=1,
            )

            grid = al.Grid.from_mask(mask=mask)

            sparse_to_grid = al.SparseToGrid.from_grid_and_unmasked_2d_grid_shape(
                unmasked_sparse_shape=(3, 3), grid=grid
            )

            assert (
                sparse_to_grid.mask_1d_index_to_sparse_1d_index
                == np.array([0, 1, 2, 3, 4])
            ).all()
            assert (
                sparse_to_grid.sparse
                == np.array(
                    [[1.0, 0.0], [0.0, -1.0], [0.0, 0.0], [0.0, 1.0], [-1.0, 0.0]]
                )
            ).all()

        def test__same_as_above_but_4x3_grid_and_mask(self):
            mask = al.Mask(
                array=np.array(
                    [
                        [True, False, True],
                        [False, False, False],
                        [False, False, False],
                        [True, False, True],
                    ]
                ),
                pixel_scale=1.0,
                sub_size=1,
            )

            grid = al.Grid.from_mask(mask=mask)

            sparse_to_grid = al.SparseToGrid.from_grid_and_unmasked_2d_grid_shape(
                unmasked_sparse_shape=(4, 3), grid=grid
            )

            assert (
                sparse_to_grid.mask_1d_index_to_sparse_1d_index
                == np.array([0, 1, 2, 3, 4, 5, 6, 7])
            ).all()
            assert (
                sparse_to_grid.sparse
                == np.array(
                    [
                        [1.5, 0.0],
                        [0.5, -1.0],
                        [0.5, 0.0],
                        [0.5, 1.0],
                        [-0.5, -1.0],
                        [-0.5, 0.0],
                        [-0.5, 1.0],
                        [-1.5, 0.0],
                    ]
                )
            ).all()

        def test__same_as_above_but_3x4_grid_and_mask(self):
            mask = al.Mask(
                array=np.array(
                    [
                        [True, False, True, True],
                        [False, False, False, False],
                        [True, False, True, True],
                    ]
                ),
                pixel_scale=1.0,
                sub_size=1,
            )

            grid = al.Grid.from_mask(mask=mask)

            sparse_to_grid = al.SparseToGrid.from_grid_and_unmasked_2d_grid_shape(
                unmasked_sparse_shape=(3, 4), grid=grid
            )

            assert (
                sparse_to_grid.mask_1d_index_to_sparse_1d_index
                == np.array([0, 1, 2, 3, 4, 5])
            ).all()
            assert (
                sparse_to_grid.sparse
                == np.array(
                    [
                        [1.0, -0.5],
                        [0.0, -1.5],
                        [0.0, -0.5],
                        [0.0, 0.5],
                        [0.0, 1.5],
                        [-1.0, -0.5],
                    ]
                )
            ).all()

        def test__mask_with_offset_centre__origin_of_sparse_to_grid_moves_to_give_same_pairings(
            self
        ):
            mask = al.Mask(
                array=np.array(
                    [
                        [True, True, True, False, True],
                        [True, True, False, False, False],
                        [True, True, True, False, True],
                        [True, True, True, True, True],
                        [True, True, True, True, True],
                    ]
                ),
                pixel_scale=1.0,
                sub_size=1,
            )

            grid = al.Grid.from_mask(mask=mask)

            # Without a change in origin, only the central 3 pixels are paired as the unmasked sparse grid overlaps
            # the central (3x3) pixels only.

            sparse_to_grid = al.SparseToGrid.from_grid_and_unmasked_2d_grid_shape(
                unmasked_sparse_shape=(3, 3), grid=grid
            )

            assert (
                sparse_to_grid.mask_1d_index_to_sparse_1d_index
                == np.array([0, 1, 2, 3, 4])
            ).all()
            assert (
                sparse_to_grid.sparse
                == np.array(
                    [[2.0, 1.0], [1.0, 0.0], [1.0, 1.0], [1.0, 2.0], [0.0, 1.0]]
                )
            ).all()

        def test__same_as_above_but_different_offset(self):
            mask = al.Mask(
                array=np.array(
                    [
                        [True, True, True, True, True],
                        [True, True, True, False, True],
                        [True, True, False, False, False],
                        [True, True, True, False, True],
                        [True, True, True, True, True],
                    ]
                ),
                pixel_scale=2.0,
                sub_size=1,
            )

            grid = al.Grid.from_mask(mask=mask)

            # Without a change in origin, only the central 3 pixels are paired as the unmasked sparse grid overlaps
            # the central (3x3) pixels only.

            sparse_to_grid = al.SparseToGrid.from_grid_and_unmasked_2d_grid_shape(
                unmasked_sparse_shape=(3, 3), grid=grid
            )

            assert (
                sparse_to_grid.mask_1d_index_to_sparse_1d_index
                == np.array([0, 1, 2, 3, 4])
            ).all()
            assert (
                sparse_to_grid.sparse
                == np.array(
                    [[2.0, 2.0], [0.0, 0.0], [0.0, 2.0], [0.0, 4.0], [-2.0, 2.0]]
                )
            ).all()

        def test__from_grid_and_unmasked_shape__sets_up_with_correct_shape_and_pixel_scales(
            self, mask_7x7
        ):
            grid = al.Grid.from_mask(mask=mask_7x7)

            sparse_to_grid = al.SparseToGrid.from_grid_and_unmasked_2d_grid_shape(
                grid=grid, unmasked_sparse_shape=(3, 3)
            )

            assert (
                sparse_to_grid.mask_1d_index_to_sparse_1d_index
                == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
            ).all()
            assert (
                sparse_to_grid.sparse
                == np.array(
                    [
                        [1.0, -1.0],
                        [1.0, 0.0],
                        [1.0, 1.0],
                        [0.0, -1.0],
                        [0.0, 0.0],
                        [0.0, 1.0],
                        [-1.0, -1.0],
                        [-1.0, 0.0],
                        [-1.0, 1.0],
                    ]
                )
            ).all()

        def test__same_as_above__but_4x3_image(self):
            mask = al.Mask(
                array=np.array(
                    [
                        [True, False, True],
                        [False, False, False],
                        [False, False, False],
                        [True, False, True],
                    ]
                ),
                pixel_scale=1.0,
                sub_size=1,
            )

            grid = al.Grid.from_mask(mask=mask)

            sparse_to_grid = al.SparseToGrid.from_grid_and_unmasked_2d_grid_shape(
                unmasked_sparse_shape=(4, 3), grid=grid
            )

            assert (
                sparse_to_grid.mask_1d_index_to_sparse_1d_index
                == np.array([0, 1, 2, 3, 4, 5, 6, 7])
            ).all()
            assert (
                sparse_to_grid.sparse
                == np.array(
                    [
                        [1.5, 0.0],
                        [0.5, -1.0],
                        [0.5, 0.0],
                        [0.5, 1.0],
                        [-0.5, -1.0],
                        [-0.5, 0.0],
                        [-0.5, 1.0],
                        [-1.5, 0.0],
                    ]
                )
            ).all()

        def test__same_as_above__but_3x4_image(self):
            mask = al.Mask(
                array=np.array(
                    [
                        [True, False, True, True],
                        [False, False, False, False],
                        [True, False, True, True],
                    ]
                ),
                pixel_scale=1.0,
                sub_size=1,
            )

            grid = al.Grid.from_mask(mask=mask)

            sparse_to_grid = al.SparseToGrid.from_grid_and_unmasked_2d_grid_shape(
                unmasked_sparse_shape=(3, 4), grid=grid
            )

            assert (
                sparse_to_grid.mask_1d_index_to_sparse_1d_index
                == np.array([0, 1, 2, 3, 4, 5])
            ).all()
            assert (
                sparse_to_grid.sparse
                == np.array(
                    [
                        [1.0, -0.5],
                        [0.0, -1.5],
                        [0.0, -0.5],
                        [0.0, 0.5],
                        [0.0, 1.5],
                        [-1.0, -0.5],
                    ]
                )
            ).all()

        def test__from_grid_and_shape__offset_mask__origin_shift_corrects(self):
            mask = al.Mask(
                array=np.array(
                    [
                        [True, True, False, False, False],
                        [True, True, False, False, False],
                        [True, True, False, False, False],
                        [True, True, True, True, True],
                        [True, True, True, True, True],
                    ]
                ),
                pixel_scale=1.0,
                sub_size=1,
            )

            grid = al.Grid.from_mask(mask=mask)

            sparse_to_grid = al.SparseToGrid.from_grid_and_unmasked_2d_grid_shape(
                unmasked_sparse_shape=(3, 3), grid=grid
            )

            assert (
                sparse_to_grid.mask_1d_index_to_sparse_1d_index
                == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
            ).all()
            assert (
                sparse_to_grid.sparse
                == np.array(
                    [
                        [2.0, 0.0],
                        [2.0, 1.0],
                        [2.0, 2.0],
                        [1.0, 0.0],
                        [1.0, 1.0],
                        [1.0, 2.0],
                        [0.0, 0.0],
                        [0.0, 1.0],
                        [0.0, 2.0],
                    ]
                )
            ).all()

    class TestUnmaskeedShapeAndWeightImage:
        def test__binned_weight_map_all_ones__kmenas_grid_is_grid_overlapping_image(
            self
        ):
            mask = al.Mask(
                array=np.array(
                    [
                        [False, False, False, False],
                        [False, False, False, False],
                        [False, False, False, False],
                        [False, False, False, False],
                    ]
                ),
                pixel_scale=0.5,
                sub_size=1,
            )

            binned_grid = al.BinnedGrid.from_mask_and_pixel_scale_binned_grid(
                mask=mask, pixel_scale_binned_grid=mask.pixel_scale
            )

            binned_weight_map = np.ones(mask.pixels_in_mask)

            sparse_to_grid_weight = al.SparseToGrid.from_total_pixels_binned_grid_and_weight_map(
                total_pixels=8,
                binned_grid=binned_grid,
                binned_weight_map=binned_weight_map,
                n_iter=10,
                max_iter=20,
                seed=1,
            )

            assert (
                sparse_to_grid_weight.sparse
                == np.array(
                    [
                        [-0.25, 0.25],
                        [0.5, -0.5],
                        [0.75, 0.5],
                        [0.25, 0.5],
                        [-0.5, -0.25],
                        [-0.5, -0.75],
                        [-0.75, 0.5],
                        [-0.25, 0.75],
                    ]
                )
            ).all()

            assert (
                sparse_to_grid_weight.mask_1d_index_to_sparse_1d_index
                == np.array([1, 1, 2, 2, 1, 1, 3, 3, 5, 4, 0, 7, 5, 4, 6, 6])
            ).all()

        def test__binned_weight_map_changes_grid_from_above(self):
            mask = al.Mask(
                array=np.array(
                    [
                        [False, False, False, False],
                        [False, False, False, False],
                        [False, False, False, False],
                        [False, False, False, False],
                    ]
                ),
                pixel_scale=0.5,
                sub_size=2,
            )

            binned_grid = al.BinnedGrid.from_mask_and_pixel_scale_binned_grid(
                mask=mask, pixel_scale_binned_grid=mask.pixel_scale
            )

            binned_weight_map = np.ones(mask.pixels_in_mask)
            binned_weight_map[0:15] = 0.00000001

            sparse_to_grid_weight = al.SparseToGrid.from_total_pixels_binned_grid_and_weight_map(
                total_pixels=8,
                binned_grid=binned_grid,
                binned_weight_map=binned_weight_map,
                n_iter=10,
                max_iter=30,
                seed=1,
            )

            assert sparse_to_grid_weight.sparse[1] == pytest.approx(
                np.array([0.4166666, -0.0833333]), 1.0e-4
            )

            assert (
                sparse_to_grid_weight.mask_1d_index_to_sparse_1d_index
                == np.array([5, 1, 0, 0, 5, 1, 1, 4, 3, 6, 7, 4, 3, 6, 2, 2])
            ).all()

        def test__binned_weight_map_all_ones__pixel_scale_binned_grid_leads_to_binning_up_by_factor_2(
            self
        ):
            mask = al.Mask(
                array=np.full(fill_value=False, shape=(8, 8)),
                pixel_scale=0.5,
                sub_size=2,
            )

            binned_grid = al.BinnedGrid.from_mask_and_pixel_scale_binned_grid(
                mask=mask, pixel_scale_binned_grid=2.0 * mask.pixel_scale
            )

            binned_weight_map = np.ones(binned_grid.shape[0])

            sparse_to_grid_weight = al.SparseToGrid.from_total_pixels_binned_grid_and_weight_map(
                total_pixels=8,
                binned_grid=binned_grid,
                binned_weight_map=binned_weight_map,
                n_iter=10,
                max_iter=30,
                seed=1,
            )

            assert (
                sparse_to_grid_weight.sparse
                == np.array(
                    [
                        [-0.5, 0.5],
                        [1.0, -1.0],
                        [1.5, 1.0],
                        [0.5, 1.0],
                        [-1.0, -0.5],
                        [-1.0, -1.5],
                        [-1.5, 1.0],
                        [-0.5, 1.5],
                    ]
                )
            ).all()

            assert (
                sparse_to_grid_weight.mask_1d_index_to_sparse_1d_index
                == np.array(
                    [
                        1,
                        1,
                        1,
                        1,
                        2,
                        2,
                        2,
                        2,
                        1,
                        1,
                        1,
                        1,
                        2,
                        2,
                        2,
                        2,
                        1,
                        1,
                        1,
                        1,
                        3,
                        3,
                        3,
                        3,
                        1,
                        1,
                        1,
                        1,
                        3,
                        3,
                        3,
                        3,
                        5,
                        5,
                        4,
                        4,
                        0,
                        0,
                        7,
                        7,
                        5,
                        5,
                        4,
                        4,
                        0,
                        0,
                        7,
                        7,
                        5,
                        5,
                        4,
                        4,
                        6,
                        6,
                        6,
                        6,
                        5,
                        5,
                        4,
                        4,
                        6,
                        6,
                        6,
                        6,
                    ]
                )
            ).all()


class TestInterpolator:
    def test_decorated_function__values_from_function_has_1_dimensions__returns_1d_result(
        self
    ):
        # noinspection PyUnusedLocal
        @al.grids.grid_interpolate
        def func(
            profile,
            grid,
            return_in_2d=False,
            return_binned=False,
            grid_radial_minimum=None,
        ):
            result = np.zeros(grid.shape[0])
            result[0] = 1
            return result

        grid = al.Grid.from_mask(
            mask=al.Mask.unmasked_from_shape_pixel_scale_and_sub_size(
                shape=(3, 3), pixel_scale=1, sub_size=1
            )
        )

        values = func(None, grid)

        assert values.ndim == 1
        assert values.shape == (9,)
        assert (values == np.array([[1, 0, 0, 0, 0, 0, 0, 0, 0]])).all()

        grid = al.Grid.from_mask(
            mask=al.Mask.unmasked_from_shape_pixel_scale_and_sub_size(
                shape=(3, 3), pixel_scale=1, sub_size=1
            )
        )
        grid.interpolator = al.Interpolator.from_mask_grid_and_pixel_scale_interpolation_grids(
            grid.mask, grid, pixel_scale_interpolation_grid=0.5
        )
        interp_values = func(None, grid)
        assert interp_values.ndim == 1
        assert interp_values.shape == (9,)
        assert (interp_values != np.array([[1, 0, 0, 0, 0, 0, 0, 0, 0]])).any()

    def test_decorated_function__values_from_function_has_2_dimensions__returns_2d_result(
        self
    ):
        # noinspection PyUnusedLocal
        @al.grids.grid_interpolate
        def func(
            profile,
            grid,
            return_in_2d=False,
            return_binned=False,
            grid_radial_minimum=None,
        ):
            result = np.zeros((grid.shape[0], 2))
            result[0, :] = 1
            return result

        grid = al.Grid.from_mask(
            mask=al.Mask.unmasked_from_shape_pixel_scale_and_sub_size(
                shape=(3, 3), pixel_scale=1, sub_size=1
            )
        )

        values = func(None, grid)

        assert values.ndim == 2
        assert values.shape == (9, 2)
        assert (
            values
            == np.array(
                [[1, 1], [0, 0], [0, 0], [0, 0], [0, 0], [0, 0], [0, 0], [0, 0], [0, 0]]
            )
        ).all()

        grid = al.Grid.from_mask(
            mask=al.Mask.unmasked_from_shape_pixel_scale_and_sub_size(
                shape=(3, 3), pixel_scale=1, sub_size=1
            )
        )
        grid.interpolator = al.Interpolator.from_mask_grid_and_pixel_scale_interpolation_grids(
            grid.mask, grid, pixel_scale_interpolation_grid=0.5
        )

        interp_values = func(None, grid)
        assert interp_values.ndim == 2
        assert interp_values.shape == (9, 2)
        assert (
            interp_values
            != np.array(
                np.array(
                    [
                        [1, 1],
                        [0, 0],
                        [0, 0],
                        [0, 0],
                        [0, 0],
                        [0, 0],
                        [0, 0],
                        [0, 0],
                        [0, 0],
                    ]
                )
            )
        ).any()

    def test__20x20_deflection_angles_no_central_pixels__interpolated_accurately(self):
        shape = (20, 20)
        pixel_scale = 1.0

        mask = al.Mask.circular_annular(
            shape=shape,
            pixel_scale=pixel_scale,
            sub_size=1,
            inner_radius_arcsec=3.0,
            outer_radius_arcsec=8.0,
        )

        grid = al.Grid.from_mask(mask=mask)

        isothermal = al.mass_profiles.SphericalIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.0
        )

        true_deflections = isothermal.deflections_from_grid(grid=grid)

        interpolator = al.Interpolator.from_mask_grid_and_pixel_scale_interpolation_grids(
            mask=mask, grid=grid, pixel_scale_interpolation_grid=1.0
        )

        interp_deflections_values = isothermal.deflections_from_grid(
            grid=interpolator.interp_grid, bypass_decorator=True
        )

        interpolated_deflections_y = interpolator.interpolated_values_from_values(
            values=interp_deflections_values[:, 0]
        )
        interpolated_deflections_x = interpolator.interpolated_values_from_values(
            values=interp_deflections_values[:, 1]
        )

        assert np.max(true_deflections[:, 0] - interpolated_deflections_y) < 0.001
        assert np.max(true_deflections[:, 1] - interpolated_deflections_x) < 0.001

    def test__move_centre_of_galaxy__interpolated_accurately(self):
        shape = (24, 24)
        pixel_scale = 1.0

        mask = al.Mask.circular_annular(
            shape=shape,
            pixel_scale=pixel_scale,
            sub_size=1,
            inner_radius_arcsec=3.0,
            outer_radius_arcsec=8.0,
            centre=(3.0, 3.0),
        )

        grid = al.Grid.from_mask(mask=mask)

        isothermal = al.mass_profiles.SphericalIsothermal(
            centre=(3.0, 3.0), einstein_radius=1.0
        )

        true_deflections = isothermal.deflections_from_grid(grid=grid)

        interpolator = al.Interpolator.from_mask_grid_and_pixel_scale_interpolation_grids(
            mask=mask, grid=grid, pixel_scale_interpolation_grid=1.0
        )

        interp_deflections_values = isothermal.deflections_from_grid(
            grid=interpolator.interp_grid, bypass_decorator=True
        )

        interpolated_deflections_y = interpolator.interpolated_values_from_values(
            values=interp_deflections_values[:, 0]
        )
        interpolated_deflections_x = interpolator.interpolated_values_from_values(
            values=interp_deflections_values[:, 1]
        )

        assert np.max(true_deflections[:, 0] - interpolated_deflections_y) < 0.001
        assert np.max(true_deflections[:, 1] - interpolated_deflections_x) < 0.001

    def test__different_interpolation_pixel_scales_still_works(self):
        shape = (28, 28)
        pixel_scale = 1.0

        mask = al.Mask.circular_annular(
            shape=shape,
            pixel_scale=pixel_scale,
            sub_size=1,
            inner_radius_arcsec=3.0,
            outer_radius_arcsec=8.0,
            centre=(3.0, 3.0),
        )

        grid = al.Grid.from_mask(mask=mask)

        isothermal = al.mass_profiles.SphericalIsothermal(
            centre=(3.0, 3.0), einstein_radius=1.0
        )

        true_deflections = isothermal.deflections_from_grid(grid=grid)

        interpolator = al.Interpolator.from_mask_grid_and_pixel_scale_interpolation_grids(
            mask=mask, grid=grid, pixel_scale_interpolation_grid=0.2
        )

        interp_deflections_values = isothermal.deflections_from_grid(
            grid=interpolator.interp_grid, bypass_decorator=True
        )

        interpolated_deflections_y = interpolator.interpolated_values_from_values(
            values=interp_deflections_values[:, 0]
        )
        interpolated_deflections_x = interpolator.interpolated_values_from_values(
            values=interp_deflections_values[:, 1]
        )

        assert np.max(true_deflections[:, 0] - interpolated_deflections_y) < 0.001
        assert np.max(true_deflections[:, 1] - interpolated_deflections_x) < 0.001

        interpolator = al.Interpolator.from_mask_grid_and_pixel_scale_interpolation_grids(
            mask=mask, grid=grid, pixel_scale_interpolation_grid=0.5
        )

        interp_deflections_values = isothermal.deflections_from_grid(
            grid=interpolator.interp_grid, bypass_decorator=True
        )

        interpolated_deflections_y = interpolator.interpolated_values_from_values(
            values=interp_deflections_values[:, 0]
        )
        interpolated_deflections_x = interpolator.interpolated_values_from_values(
            values=interp_deflections_values[:, 1]
        )

        assert np.max(true_deflections[:, 0] - interpolated_deflections_y) < 0.01
        assert np.max(true_deflections[:, 1] - interpolated_deflections_x) < 0.01

        interpolator = al.Interpolator.from_mask_grid_and_pixel_scale_interpolation_grids(
            mask=mask, grid=grid, pixel_scale_interpolation_grid=1.1
        )

        interp_deflections_values = isothermal.deflections_from_grid(
            grid=interpolator.interp_grid, bypass_decorator=True
        )

        interpolated_deflections_y = interpolator.interpolated_values_from_values(
            values=interp_deflections_values[:, 0]
        )
        interpolated_deflections_x = interpolator.interpolated_values_from_values(
            values=interp_deflections_values[:, 1]
        )

        assert np.max(true_deflections[:, 0] - interpolated_deflections_y) < 0.1
        assert np.max(true_deflections[:, 1] - interpolated_deflections_x) < 0.1



from autolens.plotters import line_yx_plotters


def plot_quantity_as_function_of_radius(
    quantity,
    radii,
    as_subplot=False,
    label=None,
    plot_axis_type="semilogy",
    effective_radius_line=None,
    einstein_radius_line=None,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    plot_legend=True,
    title="Quantity vs Radius",
    ylabel="Quantity",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    legend_fontsize=12,
    output_path=None,
    output_format="show",
    output_filename="quantity_vs_radius",
):

    vertical_lines = []
    vertical_line_labels = []

    if effective_radius_line is not None:
        vertical_lines.append(effective_radius_line)
        vertical_line_labels.append("Effective Radius")

    if einstein_radius_line is not None:
        vertical_lines.append(einstein_radius_line)
        vertical_line_labels.append("Einstein Radius")

    line_yx_plotters.plot_line(
        y=quantity,
        x=radii,
        as_subplot=as_subplot,
        label=label,
        plot_axis_type=plot_axis_type,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        plot_legend=plot_legend,
        title=title,
        ylabel=ylabel,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        legend_fontsize=legend_fontsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )



import autofit as af
import autolens as al

# In this pipeline, we'll perform a basic analysis which initialize a lens model (the lens's light, mass and source's \
# light) and then fits the source galaxy using an inversion. This pipeline uses three phases:

# Phase 1:

# Description: initialize and subtracts the lens light model.
# Lens Light: EllipticalSersic
# Lens Mass: None
# Source Light: None
# Previous Pipelines: None
# Prior Passing: None
# Notes: None

# Phase 2:

# Description: initialize the lens mass model and source light profile, using the lens subtracted image from phase 1.
# Lens Light: None
# Lens Mass: EllipitcalIsothermal + ExternalShear
# Source Light: EllipticalSersic
# Previous Pipelines: None
# Prior Passing: None
# Notes: Uses the lens light subtracted image from phase 1

# Phase 3:

# Description: Refine the lens light and mass models and source light profile, using priors from the previous 2 phases.
# Lens Light: EllipticalSersic
# Lens Mass: EllipitcalIsothermal + ExternalShear
# Source Light: EllipticalSersic
# Previous Pipelines: None
# Prior Passing: Lens light (variable -> phase 1), lens mass and source (variable -> phase 2)
# Notes: None

test_type = "full_pipeline"
test_name = "decomposed_lens_light_and_dark"
data_type = "lens_light__source_smooth"
data_resolution = "LSST"


def make_pipeline(
    name,
    phase_folders,
    pipeline_align_light_dark_centre=True,
    optimizer_class=af.MultiNest,
):

    phase1 = al.PhaseImaging(
        phase_name="phase_1__lens_sersic",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(redshift=0.5, light=al.light_profiles.EllipticalSersic)
        ),
        optimizer_class=optimizer_class,
    )

    # You'll see these lines throughout all of the example pipelines. They are used to make MultiNest sample the \
    # non-linear parameter space faster (if you haven't already, checkout the tutorial '' in howtolens/chapter_2).

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 30
    phase1.optimizer.sampling_efficiency = 0.3

    ### PHASE 2 ###

    # In phase 2, we will fit the lens galaxy's mass and source galaxy's light, where we:

    # 1) Use a lens-subtracted image generated by subtracting model lens galaxy image from phase 1.
    # 2) Initialize the priors on the centre of the lens galaxy's mass-profile by linking them to those inferred for \
    #    its light profile in phase 1.
    # 3) Have the option to use an annular mask removing the central light, if the inner_mask_radii parametr is input.

    class LensSubtractedPhase(al.PhaseImaging):
        def customize_priors(self, results):

            ## Lens Light Sersic -> Sersic ##

            self.galaxies.lens.light = results.from_phase(
                "phase_1__lens_sersic"
            ).constant.galaxies.lens.light

            ## Lens Mass, Move centre priors to centre of lens light ###

            self.galaxies.lens.mass.centre = (
                results.from_phase("phase_1__lens_sersic")
                .variable_absolute(a=0.1)
                .galaxies.lens.light.centre
            )

    phase2 = LensSubtractedPhase(
        phase_name="phase_2__lens_sie__source_sersic",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5,
                light=al.light_profiles.EllipticalSersic,
                mass=al.mass_profiles.EllipticalIsothermal,
                shear=al.mass_profiles.ExternalShear,
            ),
            source=al.GalaxyModel(
                redshift=1.0, light=al.light_profiles.EllipticalSersic
            ),
        ),
        optimizer_class=optimizer_class,
    )

    phase2.optimizer.const_efficiency_mode = False
    phase2.optimizer.n_live_points = 50
    phase2.optimizer.sampling_efficiency = 0.3

    ### PHASE 3 ###

    # In phase 3, we will fit simultaneously the lens and source galaxies, where we:

    # 1) Initialize the lens's light, mass, shear and source's light using the results of phases 1 and 2.
    # 2) Use a circular mask, to fully capture the lens and source light.

    class LensSourcePhase(al.PhaseImaging):
        def customize_priors(self, results):

            ## Lens Light, Sersic -> Sersic ###

            self.galaxies.lens.light = results.from_phase(
                "phase_1__lens_sersic"
            ).variable.galaxies.lens.light

            ## Lens Mass, SIE -> SIE, Shear -> Shear ###

            self.galaxies.lens.mass = results.from_phase(
                "phase_2__lens_sie__source_sersic"
            ).variable.galaxies.lens.mass

            self.galaxies.lens.shear = results.from_phase(
                "phase_2__lens_sie__source_sersic"
            ).variable.galaxies.lens.shear

            ### Source Light, Sersic -> Sersic ###

            self.galaxies.source = results.from_phase(
                "phase_2__lens_sie__source_sersic"
            ).variable.galaxies.source

    phase3 = LensSourcePhase(
        phase_name="phase_3__lens_sersic_sie__source_sersic",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5,
                light=al.light_profiles.EllipticalSersic,
                mass=al.mass_profiles.EllipticalIsothermal,
                shear=al.mass_profiles.ExternalShear,
            ),
            source=al.GalaxyModel(
                redshift=1.0, light=al.light_profiles.EllipticalSersic
            ),
        ),
        optimizer_class=optimizer_class,
    )

    phase3.optimizer.const_efficiency_mode = True
    phase3.optimizer.n_live_points = 75
    phase3.optimizer.sampling_efficiency = 0.3

    ### PHASE 1 ###

    # In phase 1, we will fit the lens galaxy's light and mass and one source galaxy, where we:

    # 1) Pass priors on the lens galaxy's light using the EllipticalSersic of the previous pipeline.
    # 2) Pass priors on the lens galaxy's SphericalNFW mass profile's centre using the EllipticalIsothermal fit of the
    #    previous pipeline, if the NFW centre is a free parameter.
    # 3) Pass priors on the lens galaxy's shear using the ExternalShear fit of the previous pipeline.
    # 4) Pass priors on the source galaxy's light using the EllipticalSersic of the previous pipeline.

    class LensSourcePhase(al.PhaseImaging):
        def customize_priors(self, results):

            ### Lens Light, Sersic -> Sersic ###

            self.galaxies.lens.light_mass.centre = results.from_phase(
                "phase_3__lens_sersic_sie__source_sersic"
            ).variable.galaxies.lens.light.centre

            self.galaxies.lens.light_mass.axis_ratio = results.from_phase(
                "phase_3__lens_sersic_sie__source_sersic"
            ).variable.galaxies.lens.light.axis_ratio

            self.galaxies.lens.light_mass.phi = results.from_phase(
                "phase_3__lens_sersic_sie__source_sersic"
            ).variable.galaxies.lens.light.phi

            self.galaxies.lens.light_mass.intensity = results.from_phase(
                "phase_3__lens_sersic_sie__source_sersic"
            ).variable.galaxies.lens.light.intensity

            self.galaxies.lens.light_mass.effective_radius = results.from_phase(
                "phase_3__lens_sersic_sie__source_sersic"
            ).variable.galaxies.lens.light.effective_radius

            self.galaxies.lens.light_mass.sersic_index = results.from_phase(
                "phase_3__lens_sersic_sie__source_sersic"
            ).variable.galaxies.lens.light.sersic_index

            ### Lens Mass, SIE ->  NFW ###

            if pipeline_align_light_dark_centre:

                self.galaxies.lens.dark.centre = self.galaxies.light_mass.centre

            elif not pipeline_align_light_dark_centre:

                self.galaxies.lens.dark.centre = (
                    results.from_phase("phase_3__lens_sersic_sie__source_sersic")
                    .variable_absolute(a=0.05)
                    .galaxies.lens.mass.centre
                )

            ### Lens Shear, Shear -> Shear ###

            self.galaxies.lens.shear = results.from_phase(
                "phase_3__lens_sersic_sie__source_sersic"
            ).variable.galaxies.lens.shear

            ### Source Light, Sersic -> Sersic ###

            self.galaxies.source = results.from_phase(
                "phase_3__lens_sersic_sie__source_sersic"
            ).variable.galaxies.source

    phase4 = LensSourcePhase(
        phase_name="phase_4",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5,
                light_mass=al.light_and_mass_profiles.EllipticalSersic,
                dark=al.mass_profiles.SphericalNFW,
                shear=al.mass_profiles.ExternalShear,
            ),
            source=al.GalaxyModel(
                redshift=1.0, light=al.light_profiles.EllipticalSersic
            ),
        ),
        optimizer_class=optimizer_class,
    )

    phase4.optimizer.const_efficiency_mode = True
    phase4.optimizer.n_live_points = 75
    phase4.optimizer.sampling_efficiency = 0.2

    return al.PipelineImaging(name, phase1, phase2, phase3, phase4)



import autolens as al


def test__within_radius_label_value_and_unit_string():
    string0 = al.text_util.within_radius_label_value_and_unit_string(
        prefix="mass",
        radius=1.0,
        unit_length="arcsec",
        value=30.0,
        unit_value="solMass",
        whitespace=40,
    )

    string1 = al.text_util.within_radius_label_value_and_unit_string(
        prefix="mass",
        radius=1.0,
        unit_length="arcsec",
        value=30.0,
        unit_value="solMass",
        whitespace=35,
    )

    string2 = al.text_util.within_radius_label_value_and_unit_string(
        prefix="mass",
        radius=1.0,
        unit_length="arcsec",
        value=30.0,
        unit_value="solMass",
        whitespace=30,
    )

    assert string0 == "mass_within_1.00_arcsec                 3.0000e+01 solMass"
    assert string1 == "mass_within_1.00_arcsec            3.0000e+01 solMass"
    assert string2 == "mass_within_1.00_arcsec       3.0000e+01 solMass"

    string = al.text_util.within_radius_label_value_and_unit_string(
        prefix="mass",
        radius=1.0,
        unit_length="arcsec2",
        value=40.0,
        unit_value="solMass2",
        whitespace=40,
    )

    assert string == "mass_within_1.00_arcsec2                4.0000e+01 solMass2"



import logging
import os

import numpy as np
import shutil

import autolens as al

logger = logging.getLogger(__name__)

test_data_dir = "{}/../test_files/array/".format(
    os.path.dirname(os.path.realpath(__file__))
)

test_positions_dir = "{}/../test_files/positions/".format(
    os.path.dirname(os.path.realpath(__file__))
)


class TestSignalToNoise:
    def test__image_and_noise_are_values__signal_to_noise_is_ratio_of_each(self):
        array = np.array([[1.0, 2.0], [3.0, 4.0]])

        noise = np.array([[10.0, 10.0], [30.0, 4.0]])

        data = al.AbstractData(data=array, pixel_scale=1.0, noise_map=noise)

        assert (data.signal_to_noise_map == np.array([[0.1, 0.2], [0.1, 1.0]])).all()
        assert data.signal_to_noise_max == 1.0

    def test__same_as_above__but_image_has_negative_values__replaced_with_zeros(self):
        array = np.array([[-1.0, 2.0], [3.0, -4.0]])

        noise = np.array([[10.0, 10.0], [30.0, 4.0]])

        data = al.AbstractData(data=array, pixel_scale=1.0, noise_map=noise)

        assert (data.signal_to_noise_map == np.array([[0.0, 0.2], [0.1, 0.0]])).all()
        assert data.signal_to_noise_max == 0.2


class TestAbsoluteSignalToNoise:
    def test__image_and_noise_are_values__signal_to_noise_is_absolute_image_value_over_noise(
        self
    ):
        array = np.array([[-1.0, 2.0], [3.0, -4.0]])

        noise = np.array([[10.0, 10.0], [30.0, 4.0]])

        data = al.AbstractData(data=array, pixel_scale=1.0, noise_map=noise)

        assert (
            data.absolute_signal_to_noise_map == np.array([[0.1, 0.2], [0.1, 1.0]])
        ).all()
        assert data.absolute_signal_to_noise_max == 1.0


class TestPotentialChiSquaredMap:
    def test__image_and_noise_are_values__signal_to_noise_is_absolute_image_value_over_noise(
        self
    ):
        array = np.array([[-1.0, 2.0], [3.0, -4.0]])

        noise = np.array([[10.0, 10.0], [30.0, 4.0]])

        data = al.AbstractData(data=array, pixel_scale=1.0, noise_map=noise)

        assert (
            data.potential_chi_squared_map
            == np.array([[0.1 ** 2.0, 0.2 ** 2.0], [0.1 ** 2.0, 1.0 ** 2.0]])
        ).all()
        assert data.potential_chi_squared_max == 1.0


class TestAbstractNoiseMap(object):
    class TestFromWeightMap:
        def test__weight_map_no_zeros__uses_1_over_sqrt_value(self):
            weight_map = np.array([[1.0, 4.0, 16.0], [1.0, 4.0, 16.0]])

            noise_map = al.AbstractNoiseMap.from_weight_map(
                weight_map=weight_map, pixel_scale=1.0
            )

            assert (noise_map == np.array([[1.0, 0.5, 0.25], [1.0, 0.5, 0.25]])).all()
            assert noise_map.origin == (0.0, 0.0)

        def test__weight_map_no_zeros__zeros_set_to_10000000(self):
            weight_map = np.array([[1.0, 4.0, 0.0], [1.0, 4.0, 16.0]])

            noise_map = al.AbstractNoiseMap.from_weight_map(
                weight_map=weight_map, pixel_scale=1.0
            )

            assert (noise_map == np.array([[1.0, 0.5, 1.0e8], [1.0, 0.5, 0.25]])).all()
            assert noise_map.origin == (0.0, 0.0)

    class TestFromInverseAbstractNoiseMap:
        def test__inverse_noise_map_no_zeros__uses_1_over_value(self):
            inverse_noise_map = np.array([[1.0, 4.0, 16.0], [1.0, 4.0, 16.0]])

            noise_map = al.AbstractNoiseMap.from_inverse_noise_map(
                inverse_noise_map=inverse_noise_map, pixel_scale=1.0
            )

            assert (
                noise_map == np.array([[1.0, 0.25, 0.0625], [1.0, 0.25, 0.0625]])
            ).all()
            assert noise_map.origin == (0.0, 0.0)


class TestExposureTimeMap(object):
    class TestFromExposureTimeAndBackgroundNoiseMap:
        def test__from_background_noise_map__covnerts_to_exposure_times(self):
            background_noise_map = np.array([[1.0, 4.0, 8.0], [1.0, 4.0, 8.0]])

            exposure_time_map = al.ExposureTimeMap.from_exposure_time_and_inverse_noise_map(
                pixel_scale=0.1,
                exposure_time=1.0,
                inverse_noise_map=background_noise_map,
            )

            assert (
                exposure_time_map == np.array([[0.125, 0.5, 1.0], [0.125, 0.5, 1.0]])
            ).all()
            assert exposure_time_map.origin == (0.0, 0.0)

            exposure_time_map = al.ExposureTimeMap.from_exposure_time_and_inverse_noise_map(
                pixel_scale=0.1,
                exposure_time=3.0,
                inverse_noise_map=background_noise_map,
            )

            assert (
                exposure_time_map == np.array([[0.375, 1.5, 3.0], [0.375, 1.5, 3.0]])
            ).all()
            assert exposure_time_map.origin == (0.0, 0.0)


class TestPositionsToFile(object):
    def test__load_positions__retains_list_structure(self):
        positions = al.load_positions(
            positions_path=test_positions_dir + "positions_test.dat"
        )

        assert positions == [
            [[1.0, 1.0], [2.0, 2.0]],
            [[3.0, 3.0], [4.0, 4.0], [5.0, 6.0]],
        ]

    def test__output_positions(self):
        positions = [[[4.0, 4.0], [5.0, 5.0]], [[6.0, 6.0], [7.0, 7.0], [8.0, 8.0]]]

        output_data_dir = "{}/../test_files/positions/output_test/".format(
            os.path.dirname(os.path.realpath(__file__))
        )
        if os.path.exists(output_data_dir):
            shutil.rmtree(output_data_dir)

        os.makedirs(output_data_dir)

        al.output_positions(
            positions=positions, positions_path=output_data_dir + "positions_test.dat"
        )

        positions = al.load_positions(
            positions_path=output_data_dir + "positions_test.dat"
        )

        assert positions == [
            [[4.0, 4.0], [5.0, 5.0]],
            [[6.0, 6.0], [7.0, 7.0], [8.0, 8.0]],
        ]



repeats = 1

visibilities = int(1e7)
image_pixels = int(1e5)
source_pixels = int(1e3)

shape_data = 8 * visibilities
shape_preloads = visibilities * image_pixels * 2
shape_mapping_matrix = (visibilities * source_pixels) + (image_pixels * source_pixels)

total_shape = shape_data + shape_preloads + shape_mapping_matrix

print("Data Memory Use (GB) = " + str(shape_data * 8e-9))
print("PreLoad Memory Use (GB) = " + str(shape_preloads * 8e-9))
print("Mapping Matrix Memory Use (GB) = " + str(shape_mapping_matrix * 8e-9))
print("Total Memory Use (GB) = " + str(total_shape * 8e-9))



import numpy as np
from functools import wraps

from autolens import exc
from autolens.array import grids
from autolens.array.util import array_util
from autolens.array.mapping_util import (
    array_mapping_util,
    grid_mapping_util,
    mask_mapping_util,
)
from autolens.array import scaled_array


def reshape_returned_array(func):
    @wraps(func)
    def wrapper(obj, return_in_2d=True, bypass_decorator=False, *args, **kwargs):
        """

        This wrapper decorates the _from_grid functions of profiles, which return 1D arrays of physical quantities \
        (e.g. image, convergences, potentials). Depending on the input variables, it determines whether the
        returned array is reshaped to 2D from 1D and if a sub-grid is input, it can bin the sub-gridded values to
        gridded values.

        Parameters
        ----------
        obj : autolens.model.geometry_profiles.Profile
            The profiles that owns the function
        grid : ndarray or Grid or Grid
            (y,x) in either cartesian or profiles coordinate system
        return_in_2d : bool
            If *True*, the returned array is mapped to its unmasked 2D shape, if *False* it is the masked 1D shape.
        return_binned : bool
            If *True*, the returned array which is computed on a sub-grid is binned up to the grid dimensions \
            by taking the mean of all sub-gridded values. If *False*, the array is returned on the dimensions of the \
            sub-grid.

        Returns
        -------
            An array of a physical quantity that may be in 1D or 2D and binned up from a sub-grid.
        """

        if bypass_decorator:
            return func(obj)

        grid = kwargs["grid"] if "grid" in kwargs else None
        blurring_grid = kwargs["blurring_grid"] if "blurring_grid" in kwargs else None
        psf = kwargs["psf"] if "psf" in kwargs else None
        convolver = kwargs["convolver"] if "convolver" in kwargs else None

        return_masked = kwargs["return_masked"] if "return_masked" in kwargs else True

        if hasattr(obj, "mapping"):
            mapping = obj.mapping
        elif hasattr(grid, "mapping"):
            mapping = grid.mapping
        else:
            raise exc.MappingException(
                "Unable to find mapping object from the functions input object or any of its"
                "keyword arguments."
            )

        if grid is not None and psf is not None:
            array_from_func = func(obj, grid, psf, blurring_grid)
        elif grid is not None and convolver is not None:
            array_from_func = func(obj, grid, convolver, blurring_grid)
        elif grid is not None:
            array_from_func = func(obj, grid)
        else:
            array_from_func = func(obj)

        return reshaped_array_from_array_and_mapping(
            array=array_from_func,
            mapping=mapping,
            return_in_2d=return_in_2d,
            return_masked=return_masked,
        )

    return wrapper


def reshaped_array_from_array_and_mapping(array, mapping, return_in_2d, return_masked):

    if return_in_2d and not return_masked:
        return array

    if len(array.shape) == 2:
        array_1d = mapping.array_1d_from_array_2d(array_2d=array)
    else:
        array_1d = array

    if not return_in_2d:
        return array_1d
    else:
        return mapping.scaled_array_2d_from_array_1d(array_1d=array_1d)


def reshape_returned_sub_array(func):
    @wraps(func)
    def wrapper(object, grid, *args, **kwargs):
        """

        This wrapper decorates the _from_grid functions of profiles, which return 1D arrays of physical quantities \
        (e.g. image, convergences, potentials). Depending on the input variables, it determines whether the
        returned array is reshaped to 2D from 1D and if a sub-grid is input, it can bin the sub-gridded values to
        gridded values.

        Parameters
        ----------
        object : autolens.model.geometry_profiles.Profile
            The profiles that owns the function
        grid : ndarray or Grid or Grid
            (y,x) in either cartesian or profiles coordinate system
        return_in_2d : bool
            If *True*, the returned array is mapped to its unmasked 2D shape, if *False* it is the masked 1D shape.
        return_binned : bool
            If *True*, the returned array which is computed on a sub-grid is binned up to the grid dimensions \
            by taking the mean of all sub-gridded values. If *False*, the array is returned on the dimensions of the \
            sub-grid.

        Returns
        -------
            An array of a physical quantity that may be in 1D or 2D and binned up from a sub-grid.
        """

        bypass_decorator = (
            kwargs["bypass_decorator"] if "bypass_decorator" in kwargs else False
        )

        if bypass_decorator:
            return func(object, grid)

        return_in_2d = kwargs["return_in_2d"] if "return_in_2d" in kwargs else True
        return_binned = kwargs["return_binned"] if "return_binned" in kwargs else True
        sub_array_from_func = func(object, grid)

        return reshaped_sub_array_from_sub_array_and_mapping(
            mapping=grid.mapping,
            sub_array=sub_array_from_func,
            return_in_2d=return_in_2d,
            return_binned=return_binned,
        )

    return wrapper


def reshaped_sub_array_from_sub_array_and_mapping(
    sub_array, mapping, return_in_2d, return_binned
):

    if len(sub_array.shape) == 2:
        sub_array_1d = mapping.sub_array_1d_with_sub_dimensions_from_sub_array_2d(
            sub_array_2d=sub_array
        )
    else:
        sub_array_1d = sub_array

    if not return_in_2d and not return_binned:
        return sub_array_1d

    elif not return_in_2d and return_binned:

        return mapping.array_1d_binned_from_sub_array_1d(sub_array_1d=sub_array_1d)

    elif return_in_2d and not return_binned:

        return mapping.scaled_array_2d_with_sub_dimensions_from_sub_array_1d(
            sub_array_1d=sub_array_1d
        )

    elif return_in_2d and return_binned:

        return mapping.scaled_array_2d_binned_from_sub_array_1d(
            sub_array_1d=sub_array_1d
        )


def reshape_returned_grid(func):
    @wraps(func)
    def wrapper(object, grid, *args, **kwargs):
        """

        This wrapper decorates the _from_grid functions of profiles, which return 2D grids of physical quantities \
        (e.g. deflection angles). Depending on the input variables, it determines whether the
        returned grid is reshaped to 2D from 1D and if a sub-grid is input, it can bin the sub-gridded values to
        gridded values.

        Parameters
        ----------
        object : autolens.model.geometry_profiles.Profile
            The profiles that owns the function
        mapping : ndgrid or Grid or Grid
            (y,x) in either cartesian or profiles coordinate system

        Returns
        -------
            An grid of (y,x) coordinates that may be in 1D or 2D and binned up from a sub-grid.
        """

        bypass_decorator = (
            kwargs["bypass_decorator"] if "bypass_decorator" in kwargs else False
        )

        if bypass_decorator:
            return func(object, grid)

        return_in_2d = kwargs["return_in_2d"] if "return_in_2d" in kwargs else False
        return_binned = kwargs["return_binned"] if "return_binned" in kwargs else False

        mapping = grid.mapping

        grid_from_func = func(object, grid)

        if len(grid_from_func.shape) == 3:
            grid_y = reshaped_sub_array_from_sub_array_and_mapping(
                sub_array=grid_from_func[:, :, 0],
                mapping=mapping,
                return_in_2d=return_in_2d,
                return_binned=return_binned,
            )
            grid_x = reshaped_sub_array_from_sub_array_and_mapping(
                sub_array=grid_from_func[:, :, 1],
                mapping=mapping,
                return_in_2d=return_in_2d,
                return_binned=return_binned,
            )
        elif len(grid_from_func.shape) == 2:
            grid_y = reshaped_sub_array_from_sub_array_and_mapping(
                sub_array=grid_from_func[:, 0],
                mapping=mapping,
                return_in_2d=return_in_2d,
                return_binned=return_binned,
            )
            grid_x = reshaped_sub_array_from_sub_array_and_mapping(
                sub_array=grid_from_func[:, 1],
                mapping=mapping,
                return_in_2d=return_in_2d,
                return_binned=return_binned,
            )

        return grids.Grid(arr=np.stack((grid_y, grid_x), axis=-1), mask=grid.mask)

    return wrapper


class Mapping(object):
    def __init__(self, mask):
        self.mask = mask

    @property
    def shape(self):
        return self.mask.shape

    @property
    def sub_size(self):
        return self.mask.sub_size

    @property
    def pixel_scale(self):
        return self.mask.pixel_scale

    @property
    def origin(self):
        return self.mask.origin

    @property
    def mask_1d_index_tomask_index(self):
        """A 1D array of mappings between every unmasked pixel and its 2D pixel coordinates."""
        return mask_mapping_util.sub_mask_1d_index_to_submask_index_from_mask_and_sub_size(
            mask=self.mask, sub_size=1
        ).astype(
            "int"
        )

    def array_1d_from_array_2d(self, array_2d):
        """For a 2D array (e.g. an image, noise_map, etc.) map it to a masked 1D array of valuees using this mask.

        Parameters
        ----------
        array_2d : ndarray | None | float
            The 2D array to be mapped to a masked 1D array.
        """
        if array_2d is None or isinstance(array_2d, float):
            return array_2d
        return array_mapping_util.sub_array_1d_from_sub_array_2d_mask_and_sub_size(
            mask=self.mask, sub_array_2d=array_2d, sub_size=1
        )

    def array_2d_from_array_1d(self, array_1d):
        """ Map a 1D array the same dimension as the grid to its original 2D array.

        Values which were masked in the mapping_util to the 1D array are returned as zeros.

        Parameters
        -----------
        array_1d : ndarray
            The 1D array which is mapped to its masked 2D array.
        """
        return array_mapping_util.sub_array_2d_from_sub_array_1d_mask_and_sub_size(
            sub_array_1d=array_1d, mask=self.mask, sub_size=1
        )

    def scaled_array_2d_from_array_1d(self, array_1d):
        """ Map a 1D array the same dimension as the grid to its original masked 2D array and return it as a hyper \
        array.

        Parameters
        -----------
        array_1d : ndarray
            The 1D array of which is mapped to a 2D hyper array.
        """
        return scaled_array.ScaledSquarePixelArray(
            array=self.array_2d_from_array_1d(array_1d=array_1d),
            pixel_scale=self.pixel_scale,
            origin=self.origin,
        )

    def grid_2d_from_grid_1d(self, grid_1d):
        """Map a 1D grid the same dimension as the grid to its original 2D grid.

        Values which were masked in the mapping_util to the 1D array are returned as zeros.

        Parameters
        -----------
        grid_1d : ndarray
            The 1D grid which is mapped to its masked 2D array.
        """
        return grid_mapping_util.sub_grid_2d_from_sub_grid_1d_mask_and_sub_size(
            sub_grid_1d=grid_1d, mask=self.mask, sub_size=1
        )

    def grid_1d_from_grid_2d(self, grid_2d):
        """ Map a 2D grid to its masked 1D grid..

        Values which are masked in the mapping_util to the 1D grid are returned as zeros.

        Parameters
        -----------
        grid_1d : ndgrid
            The 1D grid which is mapped to its masked 2D grid.
        """
        return grid_mapping_util.sub_grid_1d_from_sub_grid_2d_mask_and_sub_size(
            sub_grid_2d=grid_2d, mask=self.mask, sub_size=1
        )

    @property
    def sub_mask_1d_index_to_submask_index(self):
        """A 1D array of mappings between every unmasked sub pixel and its 2D sub-pixel coordinates."""
        return mask_mapping_util.sub_mask_1d_index_to_submask_index_from_mask_and_sub_size(
            mask=self.mask, sub_size=self.sub_size
        ).astype(
            "int"
        )

    @property
    @array_util.Memoizer()
    def sub_mask_1d_index_to_mask_1d_index(self):
        """The mapping_util between every sub-pixel and its host pixel.

        For example:

        - sub_to_pixel[8] = 2 -  The ninth sub-pixel is within the 3rd pixel.
        - sub_to_pixel[20] = 4 -  The twenty first sub-pixel is within the 5th pixel.
        """
        return mask_mapping_util.sub_mask_1d_index_to_mask_1d_index_from_mask(
            mask=self.mask, sub_size=self.sub_size
        ).astype("int")

    def sub_array_2d_from_sub_array_1d(self, sub_array_1d):
        """ Map a 1D sub-array the same dimension as the sub-grid (e.g. including sub-pixels) to its original masked
        2D sub array.

        Parameters
        -----------
        sub_array_1d : ndarray
            The 1D sub_array which is mapped to its masked 2D sub-array.
        """
        return array_mapping_util.sub_array_2d_from_sub_array_1d_mask_and_sub_size(
            sub_array_1d=sub_array_1d, mask=self.mask, sub_size=self.sub_size
        )

    def scaled_array_2d_with_sub_dimensions_from_sub_array_1d(self, sub_array_1d):
        """ Map a 1D sub-array the same dimension as the sub-grid to its original masked 2D sub-array and return it as
        a hyper array.

        Parameters
        -----------
        sub_array_1d : ndarray
            The 1D sub-array of which is mapped to a 2D hyper sub-array the dimensions.
        """
        return scaled_array.ScaledSquarePixelArray(
            array=self.sub_array_2d_from_sub_array_1d(sub_array_1d=sub_array_1d),
            pixel_scale=self.pixel_scale / self.sub_size,
            origin=self.origin,
        )

    def array_1d_binned_from_sub_array_1d(self, sub_array_1d):
        """For an input 1D sub-array, map its values to a 1D array of values by summing each set \of sub-pixel \
        values and dividing by the total number of sub-pixels.

        Parameters
        -----------
        sub_array_1d : ndarray
            A 1D sub-array of values (e.g. image, convergence, potential) which is mapped to
            a 1d array.
        """

        sub_length = int(self.sub_size ** 2.0)
        sub_fraction = 1.0 / sub_length

        return np.multiply(
            sub_fraction, sub_array_1d.reshape(-1, sub_length).sum(axis=1)
        )

    def scaled_array_2d_binned_from_sub_array_1d(self, sub_array_1d):
        """ Map a 1D sub-array the same dimension as the sub-grid to its original masked 2D sub-array and return it as
        a hyper array.

        Parameters
        -----------
        sub_array_1d : ndarray
            The 1D sub-array of which is mapped to a 2D hyper sub-array the dimensions.
        """

        array_1d = self.array_1d_binned_from_sub_array_1d(sub_array_1d=sub_array_1d)

        return scaled_array.ScaledSquarePixelArray(
            array=self.array_2d_from_array_1d(array_1d=array_1d),
            pixel_scale=self.pixel_scale,
            origin=self.origin,
        )

    def sub_array_1d_with_sub_dimensions_from_sub_array_2d(self, sub_array_2d):
        """ Map a 2D sub-array to its masked 1D sub-array.

        Values which are masked in the mapping_util to the 1D array are returned as zeros.

        Parameters
        -----------
        su_array_2d : ndarray
            The 2D sub-array which is mapped to its masked 1D sub-array.
        """
        return array_mapping_util.sub_array_1d_from_sub_array_2d_mask_and_sub_size(
            sub_array_2d=sub_array_2d, mask=self.mask, sub_size=self.sub_size
        )

    def sub_grid_1d_with_sub_dimensions_from_sub_grid_2d(self, sub_grid_2d):
        """For an input 1D sub-array, map its values to a 1D array of values by summing each set \of sub-pixel \
        values and dividing by the total number of sub-pixels.

        Parameters
        -----------
        sub_grid_2d : ndarray
            A 1D sub-array of values (e.g. image, convergence, potential) which is mapped to
            a 1d array.
        """
        return grid_mapping_util.sub_grid_1d_from_sub_grid_2d_mask_and_sub_size(
            sub_grid_2d=sub_grid_2d, mask=self.mask, sub_size=self.sub_size
        )

    def grid_1d_binned_from_sub_grid_1d(self, sub_grid_1d):
        """For an input 1D sub-array, map its values to a 1D array of values by summing each set \of sub-pixel \
        values and dividing by the total number of sub-pixels.

        Parameters
        -----------
        sub_grid_1d : ndarray
            A 1D sub-array of values (e.g. image, convergence, potential) which is mapped to
            a 1d array.
        """

        grid_1d_y = self.array_1d_binned_from_sub_array_1d(
            sub_array_1d=sub_grid_1d[:, 0]
        )

        grid_1d_x = self.array_1d_binned_from_sub_array_1d(
            sub_array_1d=sub_grid_1d[:, 1]
        )

        return np.stack((grid_1d_y, grid_1d_x), axis=-1)

    def grid_2d_binned_from_sub_grid_1d(self, sub_grid_1d):
        """For an input 1D sub-array, map its values to a 1D array of values by summing each set \of sub-pixel \
        values and dividing by the total number of sub-pixels.

        Parameters
        -----------
        sub_grid_1d : ndarray
            A 1D sub-array of values (e.g. image, convergence, potential) which is mapped to
            a 1d array.
        """

        grid_1d = self.grid_1d_binned_from_sub_grid_1d(sub_grid_1d=sub_grid_1d)

        return self.grid_2d_from_grid_1d(grid_1d=grid_1d)

    def sub_grid_2d_with_sub_dimensions_from_sub_grid_1d(self, sub_grid_1d):
        """ Map a 1D sub-grid the same dimension as the sub-grid (e.g. including sub-pixels) to its original masked
        2D sub grid.

        Parameters
        -----------
        sub_grid_1d : ndgrid
            The 1D grid which is mapped to its masked 2D sub-grid.
        """
        return grid_mapping_util.sub_grid_2d_from_sub_grid_1d_mask_and_sub_size(
            sub_grid_1d=sub_grid_1d, mask=self.mask, sub_size=self.sub_size
        )



from autolens.array.mapping import reshape_returned_array
import numpy as np


class MockLensImagingData(object):
    def __init__(self, imaging_data, mask, grid, blurring_grid, convolver, binned_grid):

        self.imaging_data = imaging_data
        self.pixel_scale = imaging_data.pixel_scale

        self.psf = imaging_data.psf

        self.mask = mask
        self._mask_1d = self.mask.mapping.array_1d_from_array_2d(array_2d=self.mask)

        self.grid = grid
        self.grid.new_grid_with_binned_grid(binned_grid=binned_grid)
        self.sub_size = self.grid.sub_size
        self.convolver = convolver

        self._image_1d = self.mask.mapping.array_1d_from_array_2d(
            array_2d=imaging_data.image
        )
        self._noise_map_1d = self.mask.mapping.array_1d_from_array_2d(
            array_2d=imaging_data.noise_map
        )

        self.positions = None

        self.hyper_noise_map_max = None

        self.uses_cluster_inversion = False
        self.inversion_pixel_limit = 1000
        self.inversion_uses_border = True

        self.blurring_grid = blurring_grid
        self.preload_pixelization_grids_of_planes = None

    @property
    def mapping(self):
        return self.mask.mapping

    @reshape_returned_array
    def image(self, return_in_2d=True):
        return self._image_1d

    @reshape_returned_array
    def noise_map(self, return_in_2d=True):
        return self._noise_map_1d

    @reshape_returned_array
    def signal_to_noise_map(self, return_in_2d=True):
        return self._image_1d / self._noise_map_1d


class MockLensUVPlaneData(object):
    def __init__(self, uv_plane_data, mask, grid, transformer, binned_grid):

        self.uv_plane_data = uv_plane_data
        self.pixel_scale = uv_plane_data.pixel_scale

        self.mask = mask
        self._mask_1d = self.mask.mapping.array_1d_from_array_2d(array_2d=self.mask)

        self.grid = grid
        self.grid.new_grid_with_binned_grid(binned_grid=binned_grid)
        self.sub_size = self.grid.sub_size
        self.transformer = transformer

        self.positions = None

        self.hyper_noise_map_max = None

        self.uses_cluster_inversion = False
        self.inversion_pixel_limit = 1000
        self.inversion_uses_border = True

        self.preload_pixelization_grids_of_planes = None

    def visibilities(self):
        return self.uv_plane_data.visibilities

    @property
    def visibilities_mask(self):
        return np.full(fill_value=False, shape=self.uv_plane_data.uv_wavelengths.shape)

    def noise_map(self, return_x2=False):
        if not return_x2:
            return self.uv_plane_data.noise_map
        else:
            return np.stack(
                (self.uv_plane_data.noise_map, self.uv_plane_data.noise_map), axis=-1
            )

    @property
    def primary_beam(self):
        return self.uv_plane_data.primary_beam

    def signal_to_noise_map(self):
        return self.uv_plane_data.visibilities / self.uv_plane_data.noise_map



import autolens as al

import numpy as np
import pytest


class TestVisiblities(object):
    def test__real_visibilities__intensity_image_all_ones__simple_cases(self):

        uv_wavelengths = np.ones(shape=(4, 2))
        grid_radians = np.ones(shape=(1, 2))

        transformer = al.Transformer(
            uv_wavelengths=uv_wavelengths,
            grid_radians=grid_radians,
            preload_transform=False,
        )

        image_1d = np.ones(shape=(1))

        real_visibilities = transformer.real_visibilities_from_image_1d(
            image_1d=image_1d
        )

        assert (real_visibilities == np.ones(shape=4)).all()

        uv_wavelengths = np.array([[0.2, 1.0], [0.5, 1.1], [0.8, 1.2]])

        grid_radians = np.array([[0.1, 0.2], [0.3, 0.4]])

        transformer = al.Transformer(
            uv_wavelengths=uv_wavelengths,
            grid_radians=grid_radians,
            preload_transform=False,
        )

        image_1d = np.ones(shape=(2))

        real_visibilities = transformer.real_visibilities_from_image_1d(
            image_1d=image_1d
        )

        assert real_visibilities == pytest.approx(
            np.array([1.11715, 1.68257, 1.93716]), 1.0e-4
        )

    def test__real_visibilities__intensity_image_varies__simple_cases(self):

        uv_wavelengths = np.ones(shape=(4, 2))
        grid_radians = np.ones(shape=(1, 2))

        transformer = al.Transformer(
            uv_wavelengths=uv_wavelengths,
            grid_radians=grid_radians,
            preload_transform=False,
        )

        image_1d = np.array([2.0])

        real_visibilities = transformer.real_visibilities_from_image_1d(
            image_1d=image_1d
        )

        assert (real_visibilities == np.array([2.0])).all()

        uv_wavelengths = np.array([[0.2, 1.0], [0.5, 1.1], [0.8, 1.2]])

        grid_radians = np.array([[0.1, 0.2], [0.3, 0.4]])

        transformer = al.Transformer(
            uv_wavelengths=uv_wavelengths,
            grid_radians=grid_radians,
            preload_transform=False,
        )

        image_1d = np.array([3.0, 6.0])

        real_visibilities = transformer.real_visibilities_from_image_1d(
            image_1d=image_1d
        )

        assert real_visibilities == pytest.approx(
            np.array([3.91361, 7.10136, 8.717248]), 1.0e-4
        )

    def test__real_visibilities__preload_and_non_preload_give_same_answer(self):

        uv_wavelengths = np.ones(shape=(4, 2))
        grid_radians = np.ones(shape=(1, 2))

        transformer_preload = al.Transformer(
            uv_wavelengths=uv_wavelengths,
            grid_radians=grid_radians,
            preload_transform=True,
        )
        transformer = al.Transformer(
            uv_wavelengths=uv_wavelengths,
            grid_radians=grid_radians,
            preload_transform=False,
        )

        image_1d = np.array([2.0])

        real_visibilities_via_preload = transformer_preload.real_visibilities_from_image_1d(
            image_1d=image_1d
        )
        real_visibilities = transformer.real_visibilities_from_image_1d(
            image_1d=image_1d
        )

        assert (real_visibilities_via_preload == real_visibilities).all()

    def test__imaginary_visibilities__intensity_image_all_ones__simple_cases(self):

        uv_wavelengths = np.ones(shape=(4, 2))
        grid_radians = np.ones(shape=(1, 2))

        transformer = al.Transformer(
            uv_wavelengths=uv_wavelengths,
            grid_radians=grid_radians,
            preload_transform=False,
        )

        image_1d = np.ones(shape=(1))

        imaginary_visibilities = transformer.imaginary_visibilities_from_image_1d(
            image_1d=image_1d
        )

        assert (imaginary_visibilities == np.zeros(shape=4)).all()

        uv_wavelengths = np.array([[0.2, 1.0], [0.5, 1.1], [0.8, 1.2]])

        grid_radians = np.array([[0.1, 0.2], [0.3, 0.4]])

        transformer = al.Transformer(
            uv_wavelengths=uv_wavelengths,
            grid_radians=grid_radians,
            preload_transform=False,
        )

        image_1d = np.ones(shape=(2))

        imaginary_visibilities = transformer.imaginary_visibilities_from_image_1d(
            image_1d=image_1d
        )

        assert imaginary_visibilities == pytest.approx(
            np.array([1.350411, 0.791759, 0.0]), 1.0e-4
        )

    def test__imaginary_visibilities__intensity_image_varies__simple_cases(self):

        uv_wavelengths = np.ones(shape=(4, 2))
        grid_radians = np.ones(shape=(1, 2))

        transformer = al.Transformer(
            uv_wavelengths=uv_wavelengths,
            grid_radians=grid_radians,
            preload_transform=False,
        )

        image_1d = np.array([2.0])

        imaginary_visibilities = transformer.imaginary_visibilities_from_image_1d(
            image_1d=image_1d
        )

        assert (imaginary_visibilities == np.array([0.0])).all()

        uv_wavelengths = np.array([[0.2, 1.0], [0.5, 1.1], [0.8, 1.2]])

        grid_radians = np.array([[0.1, 0.2], [0.3, 0.4]])

        transformer = al.Transformer(
            uv_wavelengths=uv_wavelengths,
            grid_radians=grid_radians,
            preload_transform=False,
        )

        image_1d = np.array([3.0, 6.0])

        imaginary_visibilities = transformer.imaginary_visibilities_from_image_1d(
            image_1d=image_1d
        )

        assert imaginary_visibilities == pytest.approx(
            np.array([6.9980971, 4.56218, 0.746069]), 1.0e-4
        )

    def test__imaginary_visibilities__preload_and_non_preload_give_same_answer(self):

        uv_wavelengths = np.ones(shape=(4, 2))
        grid_radians = np.ones(shape=(1, 2))

        transformer_preload = al.Transformer(
            uv_wavelengths=uv_wavelengths,
            grid_radians=grid_radians,
            preload_transform=True,
        )
        transformer = al.Transformer(
            uv_wavelengths=uv_wavelengths,
            grid_radians=grid_radians,
            preload_transform=False,
        )

        image_1d = np.array([2.0])

        imaginary_visibilities_via_preload = transformer_preload.imaginary_visibilities_from_image_1d(
            image_1d=image_1d
        )
        imaginary_visibilities = transformer.imaginary_visibilities_from_image_1d(
            image_1d=image_1d
        )

        assert (imaginary_visibilities_via_preload == imaginary_visibilities).all()

    def test__visiblities_from_image__same_as_individual_calculations_above(self):

        uv_wavelengths = np.array([[0.2, 1.0], [0.5, 1.1], [0.8, 1.2]])

        grid_radians = np.array([[0.1, 0.2], [0.3, 0.4]])

        transformer = al.Transformer(
            uv_wavelengths=uv_wavelengths,
            grid_radians=grid_radians,
            preload_transform=False,
        )

        image_1d = np.array([3.0, 6.0])

        visibilities = transformer.visibilities_from_image_1d(image_1d=image_1d)

        assert visibilities[:, 0] == pytest.approx(
            np.array([3.91361, 7.10136, 8.717248]), 1.0e-4
        )
        assert visibilities[:, 1] == pytest.approx(
            np.array([6.9980971, 4.56218, 0.746069]), 1.0e-4
        )

        real_visibilities = transformer.real_visibilities_from_image_1d(
            image_1d=image_1d
        )
        imaginary_visibilities = transformer.imaginary_visibilities_from_image_1d(
            image_1d=image_1d
        )

        assert (visibilities[:, 0] == real_visibilities).all()
        assert (visibilities[:, 1] == imaginary_visibilities).all()
